import pytest
import numpy
import numpy as np
from operator import attrgetter

import cupy
from cupy import testing
from cupy.testing import assert_allclose
from cupy_backends.cuda.api import driver
from cupy_backends.cuda.api import runtime
import cupyx.scipy
import cupyx.scipy.interpolate  # NOQA

try:
    import scipy
    from scipy import interpolate as sc_interpolate  # NOQA
    from scipy import special as sc_special   # NOQA
except ImportError:
    pass

interpolate_cls = ['PPoly', 'BPoly']


@testing.with_requires("scipy")
class TestPPolyCommon:
    """Test basic functionality for PPoly and BPoly."""

    @pytest.mark.parametrize('cls', interpolate_cls)
    def test_sort_check(self, cls):
        for xp, scp in [(numpy, scipy), (cupy, cupyx.scipy)]:
            c = xp.array([[1, 4], [2, 5], [3, 6]])
            x = xp.array([0, 1, 0.5])
            cls1 = getattr(scp.interpolate, cls)
            try:
                cls1(c, x)
            except ValueError:
                pass

    @pytest.mark.parametrize('cls', interpolate_cls)
    def test_ctor_c(self, cls):
        for xp, scp in [(numpy, scipy), (cupy, cupyx.scipy)]:
            # wrong shape: `c` must be at least 2D
            cls1 = getattr(scp.interpolate, cls)
            try:
                cls1(c=xp.asarray([1, 2]),
                     x=xp.asarray([0, 1]))
            except ValueError:
                pass

    @pytest.mark.parametrize('cls', interpolate_cls)
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_extend(self, xp, scp, cls):
        # Test adding new points to the piecewise polynomial
        cls = getattr(scp.interpolate, cls)
        numpy.random.seed(1234)

        order = 3
        x = numpy.unique(numpy.r_[0, 10 * numpy.random.rand(30), 10])
        x = xp.asarray(x)
        c = 2*numpy.random.rand(order+1, len(x)-1, 2, 3) - 1
        c = xp.asarray(c)

        pp = cls(c[:, :9], x[:10])
        pp.extend(c[:, 9:], x[10:])

        pp2 = cls(c[:, 10:], x[10:])
        pp2.extend(c[:, :10], x[:10])

        pp3 = cls(c, x)

        return pp.c, pp.x, pp2.c, pp2.x, pp3.c, pp3.x

    @pytest.mark.parametrize('cls', interpolate_cls)
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_extend_diff_orders(self, xp, scp, cls):
        # Test extending polynomial with different order one
        cls = getattr(scp.interpolate, cls)
        numpy.random.seed(1234)

        x = xp.linspace(0, 1, 6)
        c = xp.asarray(numpy.random.rand(2, 5))

        x2 = xp.linspace(1, 2, 6)
        c2 = xp.asarray(numpy.random.rand(4, 5))

        pp1 = cls(c, x)
        pp2 = cls(c2, x2)

        pp_comb = cls(c, x)
        pp_comb.extend(c2, x2[1:])

        # NB. doesn't match to pp1 at the endpoint, because pp1 is not
        #     continuous with pp2 as we took random coefs.
        xi1 = xp.linspace(0, 1, 300, endpoint=False)
        xi2 = xp.linspace(1, 2, 300)

        return pp1(xi1), pp_comb(xi1), pp2(xi2), pp_comb(xi2)

    @pytest.mark.parametrize('cls', interpolate_cls)
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_extend_descending(self, xp, scp, cls):
        cls = getattr(scp.interpolate, cls)
        numpy.random.seed(0)

        order = 3
        x = numpy.sort(numpy.random.uniform(0, 10, 20))
        x = xp.asarray(x)
        c = numpy.random.rand(order + 1, x.shape[0] - 1, 2, 3)
        c = xp.asarray(c)

        p1 = cls(c[:, :9], x[:10])
        p1.extend(c[:, 9:], x[10:])

        p2 = cls(c[:, 10:], x[10:])
        p2.extend(c[:, :10], x[:10])

        return p1.c, p1.x, p2.c, p2.x

    @pytest.mark.parametrize('cls', interpolate_cls)
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_shape(self, xp, scp, cls):
        cls = getattr(scp.interpolate, cls)
        numpy.random.seed(1234)

        c = numpy.random.rand(8, 12, 5, 6, 7)
        c = xp.asarray(c)
        x = numpy.sort(numpy.random.rand(13))
        x = xp.asarray(x)
        xpts = numpy.random.rand(3, 4)
        xpts = xp.asarray(xpts)

        p = cls(c, x)
        return p(xpts).shape

    @pytest.mark.parametrize('cls', interpolate_cls)
    def test_shape_2(self, cls):
        for xp, scp in [(numpy, scipy), (cupy, cupyx.scipy)]:
            cls1 = getattr(scp.interpolate, cls)
            numpy.random.seed(1234)

            c = numpy.random.rand(8, 12, 5, 6, 7)
            c = xp.asarray(c)
            x = numpy.sort(numpy.random.rand(13))
            x = xp.asarray(x)

            # 'scalars'
            p = cls1(c[..., 0, 0, 0], x)

            assert p(0.5).shape == ()
            assert p(xp.array(0.5)).shape == ()

            with pytest.raises(ValueError):
                xxx = xp.array([[0.1, 0.2], [0.4]], dtype=object)
                p(xxx)   # raises ValueError

    @pytest.mark.parametrize('cls', interpolate_cls)
    @pytest.mark.xfail(
        runtime.is_hip and driver.get_build_version() < 5_00_00000,
        reason='Fails on ROCm 4.3')
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_complex_coef(self, xp, scp, cls):
        cls = getattr(scp.interpolate, cls)
        numpy.random.seed(12345)
        x = numpy.sort(numpy.random.random(13))
        x = xp.array(x)
        c = numpy.random.random((8, 12)) * (1. + 0.3j)
        c = xp.array(c)
        # c_re, c_im = c.real, c.imag
        xpt = xp.array(numpy.random.random(5))

        p = cls(c, x)
        return [p(xpt, nu) for nu in [0, 1, 2]]

    @pytest.mark.parametrize('cls', interpolate_cls)
    @pytest.mark.parametrize('axis', [0, 1, 2, 3])
    def test_axis(self, cls, axis):
        for xp, scp in [(numpy, scipy), (cupy, cupyx.scipy)]:
            cls1 = getattr(scp.interpolate, cls)
            numpy.random.seed(12345)
            c = numpy.random.rand(3, 4, 5, 6, 7, 8)
            c = xp.asarray(c)
            c_s = c.shape
            xpt = numpy.random.random((1, 2))
            xpt = xp.asarray(xpt)

            m = c.shape[axis+1]
            x = numpy.sort(numpy.random.rand(m+1))
            x = xp.asarray(x)

            p = cls1(c, x, axis=axis)
            assert (p.c.shape ==
                    c_s[axis:axis+2] + c_s[:axis] + c_s[axis+2:])

            res = p(xpt)
            targ_shape = c_s[:axis] + xpt.shape + c_s[2+axis:]
            assert res.shape == targ_shape

            # deriv/antideriv does not drop the axis
            for p1 in [cls1(c, x, axis=axis).derivative(),
                       cls1(c, x, axis=axis).derivative(2),
                       cls1(c, x, axis=axis).antiderivative(),
                       cls1(c, x, axis=axis).antiderivative(2)]:
                assert p1.axis == p.axis

    @pytest.mark.parametrize('cls', interpolate_cls)
    @pytest.mark.parametrize('axis', [-1, 4, 5, 6])
    @testing.numpy_cupy_allclose(scipy_name='scp', accept_error=ValueError)
    def test_axis_2(self, xp, scp, cls, axis):
        cls = getattr(scp.interpolate, cls)
        numpy.random.seed(12345)
        c = numpy.random.rand(3, 4, 5, 6, 7, 8)
        c = xp.asarray(c)

        x = numpy.sort(numpy.random.rand(c.shape[-1]))
        x = xp.asarray(x)

        # c array needs two axes for the coefficients and intervals, so
        # 0 <= axis < c.ndim-1; raise otherwise
        instance = cls(c=c, x=x, axis=axis)
        return instance.c, instance.x, instance.axis


@testing.with_requires("scipy")
class TestPPoly:
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_simple(self, xp, scp):
        c = xp.array([[1, 4], [2, 5], [3, 6]])
        x = xp.array([0, 0.5, 1])
        p = scp.interpolate.PPoly(c, x)
        # testing.assert_allclose(p(0.3), 1*0.3**2 + 2*0.3 + 3)
        # testing.assert_allclose(p(0.7), 4*(0.7-0.5)**2 + 5*(0.7-0.5) + 6)
        return p(0.3), p(0.7)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_periodic(self, xp, scp):
        c = xp.array([[1, 4], [2, 5], [3, 6]])
        x = xp.array([0, 0.5, 1])
        p = scp.interpolate.PPoly(c, x, extrapolate='periodic')
        return p(1.3), p(-0.3), p(1.3, 1), p(-0.3, 1)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_read_only(self, xp, scp):
        c = xp.array([[1, 4], [2, 5], [3, 6]])
        x = xp.array([0, 0.5, 1])
        xnew = xp.array([0, 0.1, 0.2])
        scp.interpolate.PPoly(c, x, extrapolate='periodic')

        lst = []
        for writeable in (True, False):
            x.flags.writeable = writeable
            f = scp.interpolate.PPoly(c, x)
            vals = f(xnew)
            assert xp.isfinite(vals).all()
            lst.append(vals)
        return vals

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_multi_shape(self, xp, scp):
        c = numpy.random.rand(6, 2, 1, 2, 3)
        c = xp.asarray(c)
        x = xp.array([0, 0.5, 1])
        p = scp.interpolate.PPoly(c, x)
        assert p.x.shape == x.shape
        assert p.c.shape == c.shape
        assert p(0.3).shape == c.shape[2:]

        assert p(xp.random.rand(5, 6)).shape == (5, 6) + c.shape[2:]

        dp = p.derivative()
        assert dp.c.shape == (5, 2, 1, 2, 3)
        ip = p.antiderivative()
        assert ip.c.shape == (7, 2, 1, 2, 3)
        return True

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_construct_fast(self, xp, scp):
        c = xp.array([[1, 4], [2, 5], [3, 6]], dtype=float)
        x = xp.array([0, 0.5, 1])
        p = scp.interpolate.PPoly.construct_fast(c, x)
        # testing.assert_allclose(p(0.3), 1*0.3**2 + 2*0.3 + 3)
        # testing.assert_allclose(p(0.7), 4*(0.7-0.5)**2 + 5*(0.7-0.5) + 6)
        return p(0.3), p(0.7)

    def test_from_spline(self):
        numpy.random.seed(1234)
        x = numpy.sort(numpy.r_[0, numpy.random.rand(11), 1])
        y = numpy.random.rand(len(x))

        spl = sc_interpolate.splrep(x, y, s=0)
        spl = (cupy.asarray(spl[0]), cupy.asarray(spl[1]), spl[2])
        sc_spl = tuple([x.get() if isinstance(x, cupy.ndarray) else x
                        for x in spl])

        pp = cupyx.scipy.interpolate.PPoly.from_spline(spl)

        xi = np.linspace(0, 1, 200)
        testing.assert_allclose(
            pp(cupy.asarray(xi)), sc_interpolate.splev(xi, sc_spl))

        # make sure .from_spline accepts BSpline objects
        b = cupyx.scipy.interpolate.BSpline(*spl)
        ppp = cupyx.scipy.interpolate.PPoly.from_spline(b)
        testing.assert_allclose(ppp(xi), b(xi))

        # BSpline's extrapolate attribute propagates unless overridden
        t, c, k = spl
        for extrap in (None, True, False):
            b = cupyx.scipy.interpolate.BSpline(t, c, k, extrapolate=extrap)
            p = cupyx.scipy.interpolate.PPoly.from_spline(b)
            assert p.extrapolate == b.extrapolate

    def test_derivative_simple(self):
        c = cupy.array([[4, 3, 2, 1]]).T
        dc = cupy.array([[3*4, 2*3, 2]]).T
        ddc = cupy.array([[2*3*4, 1*2*3]]).T
        x = cupy.array([0, 1])

        pp = cupyx.scipy.interpolate.PPoly(c, x)
        dpp = cupyx.scipy.interpolate.PPoly(dc, x)
        ddpp = cupyx.scipy.interpolate.PPoly(ddc, x)

        testing.assert_allclose(pp.derivative().c, dpp.c)
        testing.assert_allclose(pp.derivative(2).c, ddpp.c)

    def test_derivative_eval(self):
        numpy.random.seed(1234)
        x = numpy.sort(numpy.r_[0, numpy.random.rand(11), 1])
        y = numpy.random.rand(len(x))

        spl = sc_interpolate.splrep(x, y, s=0)
        spl_cupy = (cupy.asarray(spl[0]), cupy.asarray(spl[1]), spl[2])
        pp = cupyx.scipy.interpolate.PPoly.from_spline(spl_cupy)

        xi = cupy.linspace(0, 1, 200)
        for dx in range(0, 3):
            testing.assert_allclose(pp(xi, dx),
                                    sc_interpolate.splev(xi.get(), spl, dx))

    def test_derivative(self):
        numpy.random.seed(1234)
        x = numpy.sort(numpy.r_[0, numpy.random.rand(11), 1])
        y = numpy.random.rand(len(x))

        spl = sc_interpolate.splrep(x, y, s=0, k=5)
        spl_cupy = (cupy.asarray(spl[0]), cupy.asarray(spl[1]), spl[2])
        pp = cupyx.scipy.interpolate.PPoly.from_spline(spl_cupy)

        xi = cupy.linspace(0, 1, 200)
        for dx in range(0, 10):
            testing.assert_allclose(pp(xi, dx), pp.derivative(dx)(xi),
                                    err_msg="dx=%d" % (dx,))

    def test_antiderivative_of_constant(self):
        # https://github.com/scipy/scipy/issues/4216
        PPoly = cupyx.scipy.interpolate.PPoly
        p = PPoly(cupy.asarray([[1.]]), cupy.asarray([0, 1]))
        testing.assert_allclose(p.antiderivative().c,
                                PPoly(c=cupy.asarray([[1], [0]]),
                                      x=cupy.asarray([0, 1])).c, atol=1e-15)
        testing.assert_allclose(p.antiderivative().x,
                                PPoly(c=cupy.asarray([[1], [0]]),
                                      x=cupy.asarray([0, 1])).x, atol=1e-15)

    def test_antiderivative_regression_4355(self):
        # https://github.com/scipy/scipy/issues/4355
        PPoly = cupyx.scipy.interpolate.PPoly
        p = PPoly(cupy.asarray([[1., 0.5]]), cupy.asarray([0, 1, 2]))
        q = p.antiderivative()
        testing.assert_allclose(q.c, cupy.asarray([[1, 0.5], [0, 1]]))
        testing.assert_allclose(q.x, cupy.asarray([0, 1, 2]))
        testing.assert_allclose(p.integrate(0, 2), 1.5)
        testing.assert_allclose(q(2) - q(0), 1.5)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_antiderivative_simple(self, xp, scp):
        # [ p1(x) = 3*x**2 + 2*x + 1,
        #   p2(x) = 1.6875]
        c = xp.array([[3, 2, 1], [0, 0, 1.6875]]).T
        # [ pp1(x) = x**3 + x**2 + x,
        #   pp2(x) = 1.6875*(x - 0.25) + pp1(0.25)]
        # ic = xp.array([[1, 1, 1, 0], [0, 0, 1.6875, 0.328125]]).T
        # [ ppp1(x) = (1/4)*x**4 + (1/3)*x**3 + (1/2)*x**2,
        #   ppp2(x) = (1.6875/2)*(x - 0.25)**2 + pp1(0.25)*x + ppp1(0.25)]
        # iic = xp.array([[1/4, 1/3, 1/2, 0, 0],
        #                 [0, 0, 1.6875/2, 0.328125, 0.037434895833333336]]).T
        x = xp.array([0, 0.25, 1])

        pp = scp.interpolate.PPoly(c, x)
        ipp = pp.antiderivative()
        iipp = pp.antiderivative(2)
        iipp2 = ipp.antiderivative()

        return ipp.x, ipp.c, iipp.x, iipp.c, iipp2.x, iipp2.c

    def test_antiderivative_vs_derivative(self):
        numpy.random.seed(1234)
        x = numpy.linspace(0, 1, 30)**2
        y = numpy.random.rand(len(x))
        spl = sc_interpolate.splrep(x, y, s=0, k=5)
        spl = (cupy.asarray(spl[0]), cupy.asarray(spl[1]), spl[2])
        pp = cupyx.scipy.interpolate.PPoly.from_spline(spl)

        for dx in range(0, 10):
            ipp = pp.antiderivative(dx)

            # check that derivative is inverse op
            pp2 = ipp.derivative(dx)
            assert_allclose(pp.c, pp2.c)

            # check continuity
            for k in range(dx):
                pp2 = ipp.derivative(k)

                r = 1e-13
                endpoint = r*pp2.x[:-1] + (1 - r)*pp2.x[1:]

                assert_allclose(pp2(pp2.x[1:]), pp2(endpoint),
                                rtol=1e-7, err_msg="dx=%d k=%d" % (dx, k))

    def test_antiderivative_vs_spline(self):
        numpy.random.seed(1234)
        x = numpy.sort(numpy.r_[0, numpy.random.rand(11), 1])
        y = numpy.random.rand(len(x))

        spl = sc_interpolate.splrep(x, y, s=0, k=5)
        spl_cupy = (cupy.asarray(spl[0]), cupy.asarray(spl[1]), spl[2])
        pp = cupyx.scipy.interpolate.PPoly.from_spline(spl_cupy)

        for dx in range(0, 10):
            pp2 = pp.antiderivative(dx)
            spl2 = cupyx.scipy.interpolate.splantider(spl_cupy, dx)
            spl2_np = (spl2[0].get(), spl2[1].get(), spl2[2])

            xi = cupy.linspace(0, 1, 200)
            testing.assert_allclose(pp2(xi),
                                    sc_interpolate.splev(xi.get(), spl2_np),
                                    rtol=1e-7)

    def test_antiderivative_continuity(self):
        c = cupy.array([[2, 1, 2, 2], [2, 1, 3, 3]]).T
        x = cupy.array([0, 0.5, 1])

        p = cupyx.scipy.interpolate.PPoly(c, x)
        ip = p.antiderivative()

        # check continuity
        testing.assert_allclose(ip(0.5 - 1e-9), ip(0.5 + 1e-9), rtol=1e-8)

        # check that only lowest order coefficients were changed
        p2 = ip.derivative()
        testing.assert_allclose(p2.c, p.c)

    def test_integrate(self):
        numpy.random.seed(1234)
        x = numpy.sort(numpy.r_[0, numpy.random.rand(11), 1])
        y = numpy.random.rand(len(x))

        spl = sc_interpolate.splrep(x, y, s=0, k=5)
        cp_spl = tuple([cupy.asarray(x) if isinstance(x, np.ndarray) else x
                        for x in spl])
        # spl_cupy = (cupy.asarray(spl[0]), cupy.asarray(spl[1]), spl[2])
        pp = cupyx.scipy.interpolate.PPoly.from_spline(cp_spl)

        a, b = 0.3, 0.9
        ig = pp.integrate(a, b)

        # ipp = pp.antiderivative()
        # testing.assert_allclose(ig, ipp(b) - ipp(a))
        testing.assert_allclose(ig, sc_interpolate.splint(a, b, spl))

        a, b = -0.3, 0.9
        ig = pp.integrate(a, b, extrapolate=True)
        # testing.assert_allclose(ig, ipp(b) - ipp(a))

        assert (cupy.isnan(pp.integrate(a, b, extrapolate=False)).all())

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_integrate_readonly(self, xp, scp):
        x = xp.array([1, 2, 4])
        c = xp.array([[0., 0.], [-1., -1.], [2., -0.], [1., 2.]])

        ret = []
        for writeable in (True, False):
            x.flags.writeable = writeable

            P = scp.interpolate.PPoly(c, x)
            vals = P.integrate(1, 4)
            ret.append(vals)

        return ret

    def test_integrate_periodic(self):
        x = cupy.array([1, 2, 4])
        c = cupy.array([[0., 0.], [-1., -1.], [2., -0.], [1., 2.]])

        P = cupyx.scipy.interpolate.PPoly(c, x, extrapolate='periodic')
        poly_int = P.antiderivative()

        period_int = poly_int(4) - poly_int(1)

        assert_allclose(P.integrate(1, 4), period_int)
        assert_allclose(P.integrate(-10, -7), period_int)
        assert_allclose(P.integrate(-10, -4), 2 * period_int)

        assert_allclose(P.integrate(1.5, 2.5), poly_int(2.5) - poly_int(1.5))
        assert_allclose(
            P.integrate(3.5, 5), poly_int(2) - poly_int(1) +
            poly_int(4) - poly_int(3.5))
        assert_allclose(P.integrate(3.5 + 12, 5 + 12),
                        poly_int(2) - poly_int(1) + poly_int(4) -
                        poly_int(3.5))
        assert_allclose(P.integrate(3.5, 5 + 12),
                        poly_int(2) - poly_int(1) + poly_int(4) -
                        poly_int(3.5) + 4 * period_int)

        assert_allclose(P.integrate(0, -1), poly_int(2) - poly_int(3))
        assert_allclose(P.integrate(-9, -10), poly_int(2) - poly_int(3))
        assert_allclose(
            P.integrate(0, -10), poly_int(2) - poly_int(3) - 3 * period_int)

    @pytest.mark.skip(reason='There is not an asymmetric eigenvalue solver')
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-14)
    def test_roots(self, xp, scp):
        x = xp.linspace(0, 1, 31)**2
        y = xp.sin(30*x)

        if xp is cupy:
            spl = sc_interpolate.splrep(x.get(), y.get(), s=0, k=3)
            spl = (cupy.asarray(spl[0]), cupy.asarray(spl[1]), spl[2])
        else:
            spl = sc_interpolate.splrep(x, y, s=0, k=3)

        pp = scp.interpolate.PPoly.from_spline(spl)

        r = pp.roots()
        return r

    @pytest.mark.skip(reason='There is not an asymmetric eigenvalue solver')
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-14)
    def test_roots_idzero(self, xp, scp):
        # Roots for piecewise polynomials with identically zero
        # sections.
        c = xp.array([[-1, 0.25], [0, 0], [-1, 0.25]]).T
        x = xp.array([0, 0.4, 0.6, 1.0])
        pp = scp.interpolate.PPoly(c, x)

        # ditto for p.solve(const) with sections identically equal const
        const = 2.
        c1 = c.copy()
        c1[1, :] += const
        pp1 = scp.interpolate.PPoly(c1, x)

        return pp.roots(), pp1.roots()

    @pytest.mark.skip(reason='There is not an asymmetric eigenvalue solver')
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_roots_all_zero(self, xp, scp):
        # test the code path for the polynomial being
        # identically zero everywhere
        c = xp.asarray([[0], [0]])
        x = xp.asarray([0, 1])
        p = scp.interpolate.PPoly(c, x)
        return p.roots(), p.solve(0), p.solve(1)

    @pytest.mark.skip(reason='There is not an asymmetric eigenvalue solver')
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_roots_all_zero_1(self, xp, scp):
        # test the code path for the polynomial being
        # identically zero everywhere
        c = xp.asarray([[0, 0], [0, 0]])
        x = xp.asarray([0, 1, 2])
        p = scp.interpolate.PPoly(c, x)
        return p.roots(), p.solve(0), p.solve(1)

    @pytest.mark.skip(reason='There is not an asymmetric eigenvalue solver')
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_roots_repeated(self, xp, scp):
        # Check roots repeated in multiple sections are reported only
        # once.

        # [(x + 1)**2 - 1, -x**2] ; x == 0 is a repeated root
        c = xp.array([[1, 0, -1], [-1, 0, 0]]).T
        x = xp.array([-1, 0, 1])

        pp = scp.interpolate.PPoly(c, x)
        return pp.roots(), pp.roots(extrapolate=False)

    @pytest.mark.skip(reason='There is not an asymmetric eigenvalue solver')
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_roots_discont(self, xp, scp):
        # Check that a discontinuity across zero is reported as root
        c = xp.array([[1], [-1]]).T
        x = xp.array([0, 0.5, 1])
        pp = scp.interpolate.PPoly(c, x)
        return (pp.roots(), pp.roots(discontinuity=False),
                pp.solve(0.5), pp.solve(0.5, discontinuity=False),
                pp.solve(1.5), pp.solve(1.5, discontinuity=False))

    @pytest.mark.skip(reason='There is not an asymmetric eigenvalue solver')
    def test_roots_random(self):
        # Check high-order polynomials with random coefficients
        numpy.random.seed(1234)

        num = 0

        for extrapolate in (True, False):
            for order in range(0, 20):
                x = numpy.unique(numpy.r_[0, 10 * numpy.random.rand(30), 10])
                x = cupy.asarray(x)
                c = 2*numpy.random.rand(order+1, len(x)-1, 2, 3) - 1
                c = cupy.asarray(c)

                pp = cupyx.scipy.interpolate.PPoly(c, x)
                for y in [0, numpy.random.random()]:
                    r = pp.solve(y, discontinuity=False,
                                 extrapolate=extrapolate)

                    for i in range(2):
                        for j in range(3):
                            r1 = r[i]
                            rr = r1[j]
                            if rr.size > 0:
                                # Check that the reported roots
                                # indeed are roots
                                num += rr.size
                                val = pp(rr, extrapolate=extrapolate)[:, i, j]
                                cmpval = pp(rr, nu=1,
                                            extrapolate=extrapolate)[:, i, j]
                                msg = "(%r) r = %s" % (extrapolate, repr(rr),)
                                assert_allclose((val-y) / cmpval, 0, atol=1e-7,
                                                err_msg=msg)

        # Check that we checked a number of roots
        assert num > 100, repr(num)

    @pytest.mark.skip(reason='There is not a complex root solver available')
    def test_roots_croots(self):
        # Test the complex root finding algorithm
        testing.shaped_seed(1234)

        for k in range(1, 15):
            c = testing.shaped_rand(k, 1, 130)

            if k == 3:
                # add a case with zero discriminant
                c[:, 0, 0] = 1, 2, 1

            for y in [0, testing.shaped_random()]:
                w = np.empty(c.shape, dtype=complex)
                # _ppoly._croots_poly1(c, w)

                if k == 1:
                    assert np.isnan(w).all()
                    continue

                res = 0
                cres = 0
                for i in range(k):
                    res += c[i, None] * w**(k-1-i)
                    cres += abs(c[i, None] * w**(k-1-i))
                with np.errstate(invalid='ignore'):
                    res /= cres
                res = res.ravel()
                res = res[~np.isnan(res)]
                assert_allclose(res, 0, atol=1e-10)

    @pytest.mark.parametrize('extrapolate', [True, False, None])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_extrapolate_attr(self, xp, scp, extrapolate):
        # [ 1 - x**2 ]
        c = xp.array([[-1, 0, 1]]).T
        x = xp.array([0, 1])

        pp = scp.interpolate.PPoly(c, x, extrapolate=extrapolate)
        pp_d = pp.derivative()
        pp_i = pp.antiderivative()

        xx = xp.asarray([-0.1, 1.1])

        return pp(xx), pp_i(xx), pp_d(xx)

    def binom_matrix(self, power, xp):
        n = numpy.arange(power + 1).reshape(-1, 1)
        k = numpy.arange(power + 1)
        B = sc_special.binom(n, k)
        if xp is cupy:
            B = cupy.asarray(B)
        return B[::-1, ::-1]

    def _prepare_descending(self, m, xp, scp):
        power = 3
        x = numpy.sort(numpy.random.uniform(0, 10, m + 1))
        x = xp.asarray(x)
        ca = numpy.random.uniform(-2, 2, size=(power + 1, m))
        ca = xp.asarray(ca)

        h = xp.diff(x)
        h_powers = h[None, :] ** xp.arange(power + 1)[::-1, None]
        B = self.binom_matrix(power, xp)
        cap = ca * h_powers
        cdp = xp.dot(B.T, cap)
        cd = cdp / h_powers

        pa = scp.interpolate.PPoly(ca, x, extrapolate=True)
        pd = scp.interpolate.PPoly(cd[:, ::-1], x[::-1], extrapolate=True)
        return pa, pd

    @pytest.mark.parametrize('m', [10, 20, 30])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-13)
    def test_descending(self, m, xp, scp):
        numpy.random.seed(0)
        pa, pd = self._prepare_descending(m, xp, scp)

        x_test = numpy.random.uniform(-10, 20, 100)
        x_test = xp.asarray(x_test)
        return pa(x_test), pa(x_test, 1)

    @pytest.mark.parametrize('m', [10, 20, 30])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-13)
    def test_descending_derivative(self, m, xp, scp):
        numpy.random.seed(0)
        pa, pd = self._prepare_descending(m, xp, scp)
        pa_d = pa.derivative()
        pd_d = pd.derivative()

        x_test = numpy.random.uniform(-10, 20, 100)
        x_test = xp.asarray(x_test)
        return pa_d(x_test), pd_d(x_test)

    @pytest.mark.parametrize('m', [10, 20, 30])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_descending_antiderivative(self, m, xp, scp):
        numpy.random.seed(0)
        pa, pd = self._prepare_descending(m, xp, scp)

        # Antiderivatives won't be equal because fixing continuity is
        # done in the reverse order, but surely the differences should be
        # equal.
        pa_i = pa.antiderivative()
        pd_i = pd.antiderivative()
        results = []
        for a, b in numpy.random.uniform(-10, 20, (5, 2)):
            int_a = pa.integrate(a, b)
            int_d = pd.integrate(a, b)
            results += [int_a, int_d]
            results += [pa_i(b) - pa_i(a), pd_i(b) - pd_i(a)]
        return results

    @pytest.mark.skip(reason='There is not an asymmetric eigenvalue solver')
    @pytest.mark.parametrize('m', [10, 20, 30])
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-12)
    def test_descending_roots(self, m, xp, scp):
        numpy.random.seed(0)
        pa, pd = self._prepare_descending(m, xp, scp)

        roots_d = pd.roots()
        roots_a = pa.roots()
        return roots_a, roots_d


@testing.with_requires("scipy")
class TestBPoly:
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_simple(self, xp, scp):
        x = xp.asarray([0, 1])
        c = xp.asarray([[3]])
        bp = scp.interpolate.BPoly(c, x)
        return bp(0.1)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_simple2(self, xp, scp):
        x = xp.asarray([0, 1])
        c = xp.asarray([[3], [1]])
        bp = scp.interpolate.BPoly(c, x)   # 3*(1-x) + 1*x
        # bp(0.1) = 3*0.9 + 1.*0.1
        return bp(0.1)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_simple3(self, xp, scp):
        x = xp.asarray([0, 1])
        c = xp.asarray([[3], [1], [4]])
        # 3 * (1-x)**2 + 2 * x (1-x) + 4 * x**2
        bp = scp.interpolate.BPoly(c, x)
        # bp(0.2) = 3 * 0.8*0.8 + 1 * 2*0.2*0.8 + 4 * 0.2*0.2)
        return bp(0.2)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_simple4(self, xp, scp):
        x = xp.asarray([0, 1])
        c = xp.asarray([[1], [1], [1], [2]])
        bp = scp.interpolate.BPoly(c, x)
        # bp(0.3) = 0.7**3 + 3 * 0.7**2 * 0.3 + 3 * 0.7 * 0.3**2 + 2 * 0.3**3)
        return bp(0.3)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_simple5(self, xp, scp):
        x = xp.asarray([0, 1])
        c = xp.asarray([[1], [1], [8], [2], [1]])
        bp = scp.interpolate.BPoly(c, x)
        # bp(0.3) = 0.7**4 + 4 * 0.7**3 * 0.3 + 8 * 6 * 0.7**2 * 0.3**2 +
        #           2 * 4 * 0.7 * 0.3**3 + 0.3**4)
        return bp(0.3)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_periodic(self, xp, scp):
        x = xp.asarray([0, 1, 3])
        c = xp.asarray([[3, 0], [0, 0], [0, 2]])
        # [3*(1-x)**2, 2*((x-1)/2)**2]
        bp = scp.interpolate.BPoly(c, x, extrapolate='periodic')

        # bp(3.4) = 3 * 0.6**2, bp(-1.3) = 2 * (0.7/2)**2
        # bp(3.4, 1) = -6 * 0.6, bp(-1.3, 1) = 2 * (0.7/2)
        return bp(3.4), bp(-1.3), bp(3.4, 1), bp(-1.3, 1)

    @pytest.mark.parametrize('m', [10, 20, 30])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_descending(self, xp, scp, m):
        res = []
        power = 3
        x = xp.sort(testing.shaped_random((m + 1,), xp))
        ca = testing.shaped_random((power + 1, m), xp)
        ca = ca * 2 - 1
        # We need only to flip coefficients to get it right!
        cd = ca[::-1].copy()

        pa = scp.interpolate.BPoly(ca, x, extrapolate=True)
        pd = scp.interpolate.BPoly(cd[:, ::-1], x[::-1], extrapolate=True)

        x_test = testing.shaped_random((100,), xp)
        x_test = x_test * 30 - 10

        res += [pa(x_test), pd(x_test)]
        res += [pa(x_test, 1), pd(x_test, 1)]

        pa_d = pa.derivative()
        pd_d = pd.derivative()

        res += [pa_d(x_test), pd_d(x_test)]

        # Antiderivatives won't be equal because fixing continuity is
        # done in the reverse order, but surely the differences should be
        # equal.
        pa_i = pa.antiderivative()
        pd_i = pd.antiderivative()
        points = testing.shaped_random((5, 2), xp)
        points = points * 30 - 10
        for a, b in points:
            int_a = pa.integrate(a, b)
            int_d = pd.integrate(a, b)
            res += [int_a, int_d]
            res += [pa_i(b) - pa_i(a), pd_i(b) - pd_i(a)]
        return res

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_multi_shape(self, xp, scp):
        c = testing.shaped_random((6, 2, 1, 2, 3), xp)
        x = xp.array([0, 0.5, 1])
        p = scp.interpolate.BPoly(c, x)
        x1 = testing.shaped_random((5, 6), xp)
        dp = p.derivative()
        return p(x1), dp(x1)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_interval_length(self, xp, scp):
        x = xp.asarray([0, 2])
        c = xp.asarray([[3], [1], [4]])
        bp = scp.interpolate.BPoly(c, x)
        xval = xp.asarray([0.1])
        # s = xval / 2  # s = (x - xa) / (xb - xa)
        # bp(sval) = 3 * (1-s)*(1-s) + 1 * 2*s*(1-s) + 4 * s*s
        return bp(xval)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_two_intervals(self, xp, scp):
        x = xp.asarray([0, 1, 3])
        c = xp.asarray([[3, 0], [0, 0], [0, 2]])
        bp = scp.interpolate.BPoly(c, x)  # [3*(1-x)**2, 2*((x-1)/2)**2]
        return bp(xp.asarray([0.4, 1.7]))

    @testing.numpy_cupy_allclose(scipy_name='scp')
    @pytest.mark.parametrize('extrapolate', [True, False, None])
    def test_extrapolate_attr(self, xp, scp, extrapolate):
        x = xp.asarray([0, 2])
        c = xp.asarray([[3], [1], [4]])
        x1 = xp.asarray([-0.1, 2.1])
        bp = scp.interpolate.BPoly(c, x, extrapolate=extrapolate)
        bp_d = bp.derivative()
        return bp(x1), bp_d(x1)


@testing.with_requires("scipy")
class TestBPolyCalculus:
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_derivative(self, xp, scp):
        res = []
        x = xp.asarray([0, 1, 3])
        c = xp.asarray([[3, 0], [0, 0], [0, 2]])
        # [3*(1-x)**2, 2*((x-1)/2)**2]
        bp = scp.interpolate.BPoly(c, x)
        bp_der = bp.derivative()

        # bp_der(0.4) = -6 * 0.6
        # bp_der(1.7) = 0.7
        res += [bp_der(0.4), bp_der(1.7)]

        # derivatives in-place
        # bp(0.4, nu=1) = -6 * (1 - 0.4)
        # bp(0.4, nu=2) = 6
        # bp(0.4, nu=3) = 0
        res += [bp(0.4, nu=1), bp(0.4, nu=2), bp(0.4, nu=3)]

        # bp(1.7, nu=1) = 0.7
        # bp(1.7, nu=2) = 1
        # bp(1.7, nu=3) = 0
        res += [bp(1.7, nu=1), bp(1.7, nu=2), bp(1.7, nu=3)]
        return res

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-3)
    def test_derivative_ppoly(self, xp, scp):
        # make sure it's consistent w/ power basis
        m, k = 5, 8   # number of intervals, order
        x = xp.sort(testing.shaped_random((m,), xp))
        c = testing.shaped_random((k, m-1), xp)
        bp = scp.interpolate.BPoly(c, x)
        pp = scp.interpolate.PPoly.from_bernstein_basis(bp)

        res = []
        for _ in range(k):
            bp = bp.derivative()
            pp = pp.derivative()
            xi = xp.linspace(x[0], x[-1], 21)
            res += [bp(xi), pp(xi)]
        return res

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-3)
    def test_deriv_inplace(self, xp, scp):
        m, k = 5, 8   # number of intervals, order
        x = xp.sort(testing.shaped_random((m,), xp))
        c = testing.shaped_random((k, m - 1), xp)

        # test both real and complex coefficients
        res = []
        for cc in [c.copy(), c*(1. + 2.j)]:
            # Skip for complex on older ROCm
            if (
                runtime.is_hip and
                driver.get_build_version() < 5_00_00000 and
                cc.dtype.kind == 'c'
            ):
                continue
            bp = scp.interpolate.BPoly(cc, x)
            xi = xp.linspace(x[0], x[-1], 21)
            for i in range(k):
                res += [bp(xi, i), bp.derivative(i)(xi)]
        return res

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_antiderivative_simple(self, xp, scp):
        # f(x) = x        for x \in [0, 1),
        #        (x-1)/2  for x \in [1, 3]
        #
        # antiderivative is then
        # F(x) = x**2 / 2            for x \in [0, 1),
        #        0.5*x*(x/2 - 1) + A  for x \in [1, 3]
        # where A = 3/4 for continuity at x = 1.
        x = xp.asarray([0, 1, 3])
        c = xp.asarray([[0, 0], [1, 1]])

        bp = scp.interpolate.BPoly(c, x)
        bi = bp.antiderivative()

        xx = xp.linspace(0, 3, 11)
        return bi(xx)

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=4e-6)
    def test_der_antider(self, xp, scp):
        x = xp.sort(testing.shaped_random((11,), xp))
        c = testing.shaped_random((4, 10, 2, 3), xp)
        bp = scp.interpolate.BPoly(c, x)

        xx = xp.linspace(x[0], x[-1], 100)
        return bp.antiderivative().derivative()(xx), bp(xx)

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=3e-7)
    def test_antider_ppoly(self, xp, scp):
        x = xp.sort(testing.shaped_random((11,), xp))
        c = testing.shaped_random((4, 10, 2, 3), xp)
        bp = scp.interpolate.BPoly(c, x)
        pp = scp.interpolate.PPoly.from_bernstein_basis(bp)

        xx = xp.linspace(x[0], x[-1], 10)
        return bp.antiderivative(2)(xx), pp.antiderivative(2)(xx)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_antider_continuous(self, xp, scp):
        x = xp.sort(testing.shaped_random((11,), xp))
        c = testing.shaped_random((4, 10), xp)
        bp = scp.interpolate.BPoly(c, x).antiderivative()

        xx = bp.x[1:-1]
        return bp(xx - 1e-14), bp(xx + 1e-14)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_integrate(self, xp, scp):
        x = xp.sort(testing.shaped_random((11,), xp))
        c = testing.shaped_random((4, 10), xp)
        bp = scp.interpolate.BPoly(c, x)
        pp = scp.interpolate.PPoly.from_bernstein_basis(bp)
        return bp.integrate(0, 1), pp.integrate(0, 1)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_integrate_extrap(self, xp, scp):
        c = xp.asarray([[1]])
        x = xp.asarray([0, 1])

        # default is extrapolate=True
        b = scp.interpolate.BPoly(c, x)
        # .integrate argument overrides self.extrapolate
        b1 = scp.interpolate.BPoly(c, x, extrapolate=False)

        return (b.integrate(0, 2), b1.integrate(0, 2),
                b1.integrate(0, 2, extrapolate=True))

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_integrate_periodic(self, xp, scp):
        res = []
        x = xp.asarray([1, 2, 4])
        c = xp.asarray([[0., 0.], [-1., -1.], [2., -0.], [1., 2.]])

        P = scp.interpolate.BPoly.from_power_basis(
            scp.interpolate.PPoly(c, x), extrapolate='periodic')

        # I = P.antiderivative()
        # period_int = I(4) - I(1)
        res += [
            # period_int
            P.integrate(1, 4),
            # period_int
            P.integrate(-10, -7),
            # 2 * period_int
            P.integrate(-10, -4),
            # I(2.5) - I(1.5)
            P.integrate(1.5, 2.5),
            # I(2) - I(1) + I(4) - I(3.5)
            P.integrate(3.5, 5),
            # I(2) - I(1) + I(4) - I(3.5)
            P.integrate(3.5 + 12, 5 + 12),
            # I(2) - I(1) + I(4) - I(3.5) + 4 * period_int
            P.integrate(3.5, 5 + 12),
            # I(2) - I(3)
            P.integrate(0, -1),
            # I(2) - I(3)
            P.integrate(-9, -10),
            # I(2) - I(3) - 3 * period_int
            P.integrate(0, -10)
        ]

        return res

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_antider_neg(self, xp, scp):
        # .derivative(-nu) ==> .andiderivative(nu) and vice versa
        c = xp.asarray([[1]])
        x = xp.asarray([0, 1])
        b = scp.interpolate.BPoly(c, x)

        xx = xp.linspace(0, 1, 21)

        return (b.derivative(-1)(xx), b.antiderivative()(xx),
                b.derivative(1)(xx), b.antiderivative(-1)(xx))


@testing.with_requires("scipy")
class TestPolyConversions:
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_bp_from_pp(self, xp, scp):
        x = xp.asarray([0, 1, 3])
        c = xp.asarray([[3, 2], [1, 8], [4, 3]])
        pp = scp.interpolate.PPoly(c, x)
        bp = scp.interpolate.BPoly.from_power_basis(pp)
        pp1 = scp.interpolate.PPoly.from_bernstein_basis(bp)

        x1 = xp.asarray([0.1, 1.4])
        return pp(x1), bp(x1), pp(x1), pp1(x1)

    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=6e-7)
    def test_bp_from_pp_random(self, xp, scp):
        m, k = 5, 8   # number of intervals, order
        x = xp.sort(testing.shaped_random((m,), xp))
        c = testing.shaped_random((k, m - 1), xp)
        pp = scp.interpolate.PPoly(c, x)
        bp = scp.interpolate.BPoly.from_power_basis(pp)
        pp1 = scp.interpolate.PPoly.from_bernstein_basis(bp)

        x1 = xp.linspace(x[0], x[-1], 21)
        return pp(x1), bp(x1), pp(x1), pp1(x1)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_pp_from_bp(self, xp, scp):
        x = xp.asarray([0, 1, 3])
        c = xp.asarray([[3, 3], [1, 1], [4, 2]])
        bp = scp.interpolate.BPoly(c, x)
        pp = scp.interpolate.PPoly.from_bernstein_basis(bp)
        bp1 = scp.interpolate.BPoly.from_power_basis(pp)

        x1 = xp.asarray([0.1, 1.4])
        return bp(x1), pp(x1), bp(x1), bp1(x1)

    @testing.numpy_cupy_allclose(scipy_name='scp', accept_error=True)
    @pytest.mark.parametrize('comb', [
        (attrgetter('PPoly'), attrgetter('PPoly.from_bernstein_basis')),
        (attrgetter('BPoly'), attrgetter('PPoly.from_power_basis'))])
    def test_broken_conversions(self, xp, scp, comb):
        # regression test for scipy/gh-10597: from_power_basis only
        # accepts PPoly etc.
        x = xp.asarray([0, 1, 3])
        c = xp.asarray([[3, 3], [1, 1], [4, 2]])

        get_interp_cls, get_conv_meth = comb
        pp = get_interp_cls(scp.interpolate)(c, x)
        get_conv_meth(scp.interpolate)(pp)


@testing.with_requires("scipy")
class TestBPolyFromDerivatives:
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_make_poly_1(self, xp, scp):
        c1 = scp.interpolate.BPoly._construct_from_derivatives(
            0, 1, [2], [3])
        return c1

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_make_poly_2(self, xp, scp):
        c1 = scp.interpolate.BPoly._construct_from_derivatives(
            0, 1, [1, 0], [1])
        # f'(0) = 3
        c2 = scp.interpolate.BPoly._construct_from_derivatives(
            0, 1, [2, 3], [1])
        # f'(1) = 3
        c3 = scp.interpolate.BPoly._construct_from_derivatives(
            0, 1, [2], [1, 3])
        return c1, c2, c3

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_make_poly_3(self, xp, scp):
        # f'(0)=2, f''(0)=3
        c1 = scp.interpolate.BPoly._construct_from_derivatives(
            0, 1, [1, 2, 3], [4])
        # f'(1)=2, f''(1)=3
        c2 = scp.interpolate.BPoly._construct_from_derivatives(
            0, 1, [1], [4, 2, 3])
        # f'(0)=2, f'(1)=3
        c3 = scp.interpolate.BPoly._construct_from_derivatives(
            0, 1, [1, 2], [4, 3])
        return c1, c2, c3

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_make_poly_12(self, xp, scp):
        ya = xp.r_[0, testing.shaped_random((5,), xp)]
        yb = xp.r_[0, testing.shaped_random((5,), xp)]

        c = scp.interpolate.BPoly._construct_from_derivatives(0, 1, ya, yb)
        pp = scp.interpolate.BPoly(c[:, None], xp.asarray([0, 1]))
        res = []
        for _ in range(6):
            res += [pp(0.), pp(1.)]
            pp = pp.derivative()
        return res

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_raise_degree(self, xp, scp):
        x = xp.asarray([0, 1])
        k, d = 8, 5
        c = testing.shaped_random((k, 1, 2, 3, 4), xp)
        bp = scp.interpolate.BPoly(c, x)

        c1 = scp.interpolate.BPoly._raise_degree(c, d)
        bp1 = scp.interpolate.BPoly(c1, x)

        x1 = xp.linspace(0, 1, 11)
        return bp(x1), bp1(x1)

    @testing.numpy_cupy_allclose(scipy_name='scp', accept_error=ValueError)
    def test_xi_yi(self, xp, scp):
        scp.interpolate.BPoly.from_derivatives(
            xp.asarray([0, 1]), xp.asarray([0]))

    @testing.numpy_cupy_allclose(scipy_name='scp', accept_error=ValueError)
    def test_coords_order(self, xp, scp):
        xi = xp.asarray([0, 0, 1])
        yi = xp.asarray([[0], [0], [0]])
        scp.interpolate.BPoly.from_derivatives(xi, yi)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_zeros(self, xp, scp):
        xi = xp.asarray([0, 1, 2, 3])
        # NB: will have to raise the degree
        yi = [[0, 0], [0], [0, 0], [0, 0]]
        pp = scp.interpolate.BPoly.from_derivatives(xi, yi)
        assert pp.c.shape == (4, 3)

        ppd = pp.derivative()
        res = []
        for x1 in [0., 0.1, 1., 1.1, 1.9, 2., 2.5]:
            res += [pp(x1), ppd(x1)]
        return res

    def _make_random_mk(self, m, k, xp):
        # k derivatives at each breakpoint
        xi = xp.asarray([1. * j**2 for j in range(m+1)])
        yi = [testing.shaped_random((k,), xp) for j in range(m+1)]
        return xi, yi

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_random_12(self, xp, scp):
        m, k = 5, 12
        xi, yi = self._make_random_mk(m, k, xp)
        pp = scp.interpolate.BPoly.from_derivatives(xi, yi)

        res = []
        for _ in range(k//2):
            res.append(pp(xi))
            pp = pp.derivative()
        return res

    @testing.numpy_cupy_allclose(scipy_name='scp', accept_error=ValueError)
    def test_order_zero(self, xp, scp):
        m, k = 5, 12
        xi, yi = self._make_random_mk(m, k, xp)
        scp.interpolate.BPoly.from_derivatives(xi, yi, 0)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_orders_too_high(self, xp, scp):
        m, k = 5, 12
        xi, yi = self._make_random_mk(m, k, xp)

        # this is still ok
        scp.interpolate.BPoly.from_derivatives(xi, yi, orders=2*k-1)
        with pytest.raises(ValueError):
            # but this is not
            scp.interpolate.BPoly.from_derivatives(xi, yi, orders=2*k)
        return True

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_orders_global(self, xp, scp):
        m, k = 5, 12
        xi, yi = self._make_random_mk(m, k, xp)

        # ok, this is confusing. Local polynomials will be of the order 5
        # which means that up to the 2nd derivatives will be used at each point
        order = 5
        pp = scp.interpolate.BPoly.from_derivatives(xi, yi, orders=order)

        res = []
        for _ in range(order//2+1):
            res += [pp(xi[1:-1] - 1e-12), pp(xi[1:-1] + 1e-12)]
            pp = pp.derivative()
        # assert_(not xp.allclose(pp(xi[1:-1] - 1e-12), pp(xi[1:-1] + 1e-12)))

        # now repeat with `order` being even: on each interval, it uses
        # order//2 'derivatives' @ the right-hand endpoint and
        # order//2+1 @ 'derivatives' the left-hand endpoint
        order = 6
        pp = scp.interpolate.BPoly.from_derivatives(xi, yi, orders=order)
        for _ in range(order//2):
            res += [pp(xi[1:-1] - 1e-12), pp(xi[1:-1] + 1e-12)]
            pp = pp.derivative()
        # assert_(not xp.allclose(pp(xi[1:-1] - 1e-12), pp(xi[1:-1] + 1e-12)))
        return res

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_orders_local(self, xp, scp):
        m, k = 7, 12
        xi, yi = self._make_random_mk(m, k, xp)

        orders = [o + 1 for o in range(m)]
        res = []
        for i, x in enumerate(xi[1:-1]):
            pp = scp.interpolate.BPoly.from_derivatives(xi, yi, orders=orders)
            for _ in range(orders[i] // 2 + 1):
                res += [pp(x - 1e-12), pp(x + 1e-12)]
                pp = pp.derivative()
            # assert_(not xp.allclose(pp(x - 1e-12), pp(x + 1e-12)))
        return res

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_yi_trailing_dims(self, xp, scp):
        m, k = 7, 5
        xi = xp.sort(testing.shaped_random((m+1,), xp))
        yi = testing.shaped_random((m+1, k, 6, 7, 8), xp)
        pp = scp.interpolate.BPoly.from_derivatives(xi, yi)
        return pp.c.shape

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_scipy_gh_5430(self, xp, scp):
        # At least one of these raises an error unless gh-5430 is
        # fixed. In py2k an int is implemented using a C long, so
        # which one fails depends on your system. In py3k there is only
        # one arbitrary precision integer type, so both should fail.
        res = []
        orders = xp.int32(1)
        p = scp.interpolate.BPoly.from_derivatives(
            xp.asarray([0, 1]), [[0], [0]], orders=orders)
        res.append(p(0))
        orders = xp.int64(1)
        p = scp.interpolate.BPoly.from_derivatives(
            xp.asarray([0, 1]), [[0], [0]], orders=orders)
        res.append(p(0))
        orders = 1
        # This worked before; make sure it still works
        p = scp.interpolate.BPoly.from_derivatives(
            xp.asarray([0, 1]), [[0], [0]], orders=orders)
        res.append(p(0))
        orders = 1
        return res


@testing.with_requires("scipy")
class TestNdPPoly:

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_simple_1d(self, xp, scp):
        c = testing.shaped_random((4, 5), xp)
        x = xp.linspace(0, 1, 5+1)

        xi = testing.shaped_random((200,), xp)

        p = scp.interpolate.NdPPoly(c, (x,))
        v1 = p((xi,))
        return v1

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_simple_2d(self, xp, scp):
        c = testing.shaped_random((4, 5, 6, 7), xp)
        x = xp.linspace(0, 1, 6+1)
        y = xp.linspace(0, 1, 7+1)**2

        xi = testing.shaped_random((200,), xp)
        yi = testing.shaped_random((200,), xp)

        p = scp.interpolate.NdPPoly(c, (x, y))
        result = []
        for nu in (None, (0, 0), (0, 1), (1, 0), (2, 3), (9, 2)):
            result.append(p(xp.c_[xi, yi], nu=nu))
        return result

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_simple_3d(self, xp, scp):
        c = testing.shaped_random((4, 5, 6, 7, 8, 9), xp)
        x = xp.linspace(0, 1, 7+1)
        y = xp.linspace(0, 1, 8+1)**2
        z = xp.linspace(0, 1, 9+1)**3

        xi = testing.shaped_random((40,), xp)
        yi = testing.shaped_random((40,), xp)
        zi = testing.shaped_random((40,), xp)

        p = scp.interpolate.NdPPoly(c, (x, y, z))
        result = []
        for nu in (None, (0, 0, 0), (0, 1, 0), (1, 0, 0), (2, 3, 0),
                   (6, 0, 2)):
            result.append(p((xi, yi, zi), nu=nu))
        return result

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_simple_4d(self, xp, scp):
        c = testing.shaped_random((4, 5, 6, 7, 8, 9, 10, 11), xp)
        x = xp.linspace(0, 1, 8+1)
        y = xp.linspace(0, 1, 9+1)**2
        z = xp.linspace(0, 1, 10+1)**3
        u = xp.linspace(0, 1, 11+1)**4

        xi = testing.shaped_random((20,), xp)
        yi = testing.shaped_random((20,), xp)
        zi = testing.shaped_random((20,), xp)
        ui = testing.shaped_random((20,), xp)

        p = scp.interpolate.NdPPoly(c, (x, y, z, u))
        v1 = p((xi, yi, zi, ui))
        return v1

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_deriv_1d(self, xp, scp):
        c = testing.shaped_random((4, 5), xp)
        x = xp.linspace(0, 1, 5+1)

        p = scp.interpolate.NdPPoly(c, (x,))

        # derivative
        dp = p.derivative(nu=[1])

        # antiderivative
        ip = p.antiderivative(nu=[2])
        return dp.c, ip.c

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_deriv_3d(self, xp, scp):
        c = testing.shaped_random((4, 5, 6, 7, 8, 9), xp)
        x = xp.linspace(0, 1, 7+1)
        y = xp.linspace(0, 1, 8+1)**2
        z = xp.linspace(0, 1, 9+1)**3

        p = scp.interpolate.NdPPoly(c, (x, y, z))

        # differentiate vs x
        dpx = p.derivative(nu=[2])

        # antidifferentiate vs y
        dpy = p.antiderivative(nu=[0, 1, 0])

        # differentiate vs z
        dpz = p.derivative(nu=[0, 0, 3])

        return dpx.c, dpy.c, dpz.c

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_deriv_3d_simple(self, xp, scp):
        # Integrate to obtain function x y**2 z**4 / (2! 4!)

        c = xp.ones((1, 1, 1, 3, 4, 5))
        x = xp.linspace(0, 1, 3+1)**1
        y = xp.linspace(0, 1, 4+1)**2
        z = xp.linspace(0, 1, 5+1)**3

        p = scp.interpolate.NdPPoly(c, (x, y, z))
        ip = p.antiderivative((1, 0, 4))
        ip = ip.antiderivative((0, 2, 0))

        xi = testing.shaped_random((20,), xp)
        yi = testing.shaped_random((20,), xp)
        zi = testing.shaped_random((20,), xp)

        return ip((xi, yi, zi))

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_integrate_2d(self, xp, scp):
        c = testing.shaped_random((4, 5, 16, 17), xp, dtype=xp.float64)
        x = xp.linspace(0, 1, 16+1)**1
        y = xp.linspace(0, 1, 17+1)**2

        # make continuously differentiable so that nquad() has an
        # easier time
        fix_continuity_mod = attrgetter(
            'interpolate._interpolate._fix_continuity')
        if xp is not cupy:
            fix_continuity_mod = attrgetter(
                'interpolate._ppoly.fix_continuity')

        fix_continuity = fix_continuity_mod(scp)

        c = c.transpose(0, 2, 1, 3)
        cx = c.reshape(c.shape[0], c.shape[1], -1).copy()
        fix_continuity(cx, x, 2)
        c = cx.reshape(c.shape)
        c = c.transpose(0, 2, 1, 3)
        c = c.transpose(1, 3, 0, 2)
        cx = c.reshape(c.shape[0], c.shape[1], -1).copy()
        fix_continuity(cx, y, 2)
        c = cx.reshape(c.shape)
        c = c.transpose(2, 0, 3, 1).copy()

        # Check integration
        p = scp.interpolate.NdPPoly(c, (x, y))

        result = []
        for ranges in [[(0, 1), (0, 1)],
                       [(0, 0.5), (0, 1)],
                       [(0, 1), (0, 0.5)],
                       [(0.3, 0.7), (0.6, 0.2)]]:
            ig = p.integrate(ranges)
            result.append(ig)
        return result

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_integrate_1d(self, xp, scp):
        c = testing.shaped_random((4, 5, 6, 16, 17, 18), xp)
        x = xp.linspace(0, 1, 16+1)**1
        y = xp.linspace(0, 1, 17+1)**2
        z = xp.linspace(0, 1, 18+1)**3

        # Check 1-D integration
        p = scp.interpolate.NdPPoly(c, (x, y, z))

        u = testing.shaped_random((200,), xp)
        v = testing.shaped_random((200,), xp)
        a, b = 0.2, 0.7

        result = []

        px = p.integrate_1d(a, b, axis=0)
        pax = p.antiderivative((1, 0, 0))
        result += [px((u, v)), pax((b, u, v)) - pax((a, u, v))]

        py = p.integrate_1d(a, b, axis=1)
        pay = p.antiderivative((0, 1, 0))
        result += [py((u, v)), pay((u, b, v)) - pay((u, a, v))]

        pz = p.integrate_1d(a, b, axis=2)
        paz = p.antiderivative((0, 0, 1))
        result += [pz((u, v)), paz((u, v, b)) - paz((u, v, a))]

        return result



import contextlib
from itertools import chain
from typing import Any, Iterator, Optional, Union

import numpy

from cupy._core.core import ndarray
import cupy._creation.basic as _creation_basic
import cupy._manipulation.dims as _manipulation_dims
from cupy.cuda.device import Device
from cupy.cuda.stream import Event
from cupy.cuda.stream import Stream
from cupy.cuda.stream import get_current_stream

from cupyx.distributed.array import _modes
from cupyx.distributed.array import _index_arith
from cupyx.distributed.array import _data_transfer
from cupyx.distributed.array._data_transfer import _Communicator


class _ArrayPlaceholder:
    # Mocks ndarray
    # Eventually overwritten by PartialUpdates entirely, so
    # any operation on _DataPlaceholder can be skipped
    shape: tuple[int, ...]
    device: Device

    def __init__(self, shape: tuple[int, ...], device: Device) -> None:
        self.shape = shape
        self.device = device

    def reshape(self, new_shape: tuple[int, ...]) -> '_ArrayPlaceholder':
        return _ArrayPlaceholder(new_shape, self.device)

    def to_ndarray(
            self, mode: '_modes.Mode', dtype: numpy.dtype) -> ndarray:
        with self.device:
            if mode is _modes.REPLICA:
                data = _creation_basic.empty(self.shape, dtype)
            else:
                value = mode.identity_of(dtype)
                data = _creation_basic.full(self.shape, value, dtype)

            # We avoid 0D array because we expect data[idx] to return a view
            return _manipulation_dims.atleast_1d(data)


class _Chunk:
    array: Union[ndarray, _ArrayPlaceholder]
    ready: Event
    index: tuple[slice, ...]
    updates: list[_data_transfer._PartialUpdate]
    prevent_gc: Any = None     # TODO: Release it to avoid OOM

    # Rule: whenever data is DataPlaceholder, ready is empty

    def __init__(
        self, data: Union[ndarray, _ArrayPlaceholder], ready: Event,
        index: tuple[slice, ...],
        updates: Optional[list[_data_transfer._PartialUpdate]] = None,
        prevent_gc: Any = None
    ) -> None:
        self.array = data
        self.ready = ready
        self.index = index
        self.updates = updates if updates is not None else []
        self.prevent_gc = prevent_gc

    @classmethod
    def create_placeholder(
        cls, shape: tuple[int, ...], device: Union[int, Device],
        index: tuple[slice, ...],
        updates: Optional[list[_data_transfer._PartialUpdate]] = None,
    ) -> '_Chunk':
        if isinstance(device, int):
            device = Device(device)

        data = _ArrayPlaceholder(shape, device)
        with device:
            ready = Event()
        if updates is None:
            updates = []

        return _Chunk(data, ready, index, updates)

    @contextlib.contextmanager
    def on_ready(self) -> Iterator[Stream]:
        with self.array.device:
            stream = get_current_stream()
            stream.wait_event(self.ready)
            yield stream

    def add_update(
        self, update: _data_transfer._AsyncData, idx: tuple[slice, ...],
    ) -> None:
        self.updates.append((update, idx))

    def copy(self) -> '_Chunk':
        # TODO: Calling flush here would reduce the amount of future copying
        if isinstance(self.array, _ArrayPlaceholder):
            data = self.array
            ready = self.ready
        else:
            with self.on_ready() as stream:
                data = self.array.copy()
                ready = stream.record()

        return _Chunk(data, ready, self.index, list(self.updates),
                      prevent_gc=self.prevent_gc)

    def flush(self, mode: '_modes.Mode') -> None:
        """Apply all updates in-place."""
        if len(self.updates) == 0:
            return

        if isinstance(self.array, _ArrayPlaceholder):
            dtype = self.updates[0][0].array.dtype
            self.array = self.array.to_ndarray(mode, dtype)

        with self.on_ready() as stream:
            for update_data, idx in self.updates:
                stream.wait_event(update_data.ready)
                if mode is _modes.REPLICA:
                    self.array[idx] = update_data.array
                else:
                    self.array[idx] = mode.func(
                        self.array[idx], update_data.array)

            stream.record(self.ready)
            self.prevent_gc = (self.prevent_gc, self.updates)
            self.updates = []

    def apply_to(
        self, target: '_Chunk', mode: '_modes.Mode',
        shape: tuple[int, ...],
        comms: dict[int, _data_transfer._Communicator],
        streams: dict[int, Stream],
    ) -> None:
        # Overwrite target with mode.func(self, target) on their overlaps
        # This is just appending part of self to target.updates in the mode
        src_chunk = self
        dst_chunk = target

        assert len(src_chunk.updates) == 0
        assert isinstance(src_chunk.array, ndarray)

        src_dev = src_chunk.array.device.id
        dst_dev = dst_chunk.array.device.id
        src_idx = src_chunk.index
        dst_idx = dst_chunk.index

        intersection = _index_arith._index_intersection(
            src_idx, dst_idx, shape)
        if intersection is None:
            return

        src_new_idx = _index_arith._index_for_subindex(
            src_idx, intersection, shape)
        dst_new_idx = _index_arith._index_for_subindex(
            dst_idx, intersection, shape)

        data_to_transfer = _data_transfer._AsyncData(
            src_chunk.array[src_new_idx], src_chunk.ready,
            src_chunk.prevent_gc)

        if mode is not _modes.REPLICA and not mode.idempotent:
            data_to_transfer = data_to_transfer.copy()

        update = _data_transfer._transfer(
            comms[src_dev], streams[src_dev], data_to_transfer,
            comms[dst_dev], streams[dst_dev], dst_dev)
        dst_chunk.add_update(update, dst_new_idx)

        if mode is not _modes.REPLICA and not mode.idempotent:
            dtype = src_chunk.array.dtype
            with data_to_transfer.on_ready() as stream:
                # Now src data has been copied, so we can write on src_chunk
                src_chunk.array[src_new_idx] = mode.identity_of(dtype)
                stream.record(src_chunk.ready)

    def set_identity_on_intersection(
        self, idx: tuple[slice, ...], shape: tuple[int, ...], identity,
    ) -> None:
        assert isinstance(self.array, ndarray)

        intersection = _index_arith._index_intersection(self.index, idx, shape)
        if intersection is None:
            return
        self_new_idx = _index_arith._index_for_subindex(
            self.index, intersection, shape)
        with self.on_ready() as stream:
            self.array[self_new_idx] = identity
            stream.record(self.ready)

    def set_identity_on_overwritten_entries(self, identity) -> None:
        if isinstance(self.array, _ArrayPlaceholder):
            return

        with self.on_ready() as stream:
            for _, idx in self.updates:
                self.array[idx] = identity
            stream.record(self.ready)


def _all_reduce_intersections(
    op_mode: '_modes._OpMode', shape: tuple[int, ...],
    chunk_map: dict[int, list[_Chunk]],
    comms: dict[int, _Communicator], streams: dict[int, Stream],
) -> None:
    chunks_list = list(chain.from_iterable(chunk_map.values()))

    for i in range(len(chunks_list)):
        src_chunk = chunks_list[i]
        src_chunk.flush(op_mode)

        for j in range(i + 1, len(chunks_list)):
            dst_chunk = chunks_list[j]

            src_chunk.apply_to(dst_chunk, op_mode, shape, comms, streams)

    for j in range(len(chunks_list) - 1, -1, -1):
        src_chunk = chunks_list[j]
        src_chunk.flush(_modes.REPLICA)

        for i in range(j):
            dst_chunk = chunks_list[i]
            src_chunk.apply_to(
                dst_chunk, _modes.REPLICA, shape, comms, streams)



import pytest

import cupy
from cupy import testing
import cupyx
import cupyx.scipy.special  # NOQA


atol = {
    'default': 1e-6,
    cupy.float16: 1e-2,
}
rtol = {
    'default': 1e-6,
    cupy.float16: 1e-2,
}


@testing.with_requires('scipy')
class TestSoftmax:

    @testing.for_all_dtypes(no_bool=True)
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_1dim(self, xp, scp, dtype):
        x = xp.arange(400)
        return scp.special.softmax(x)

    @testing.for_all_dtypes(no_float16=True, no_bool=True)
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=atol, rtol=rtol)
    def test_2dim(self, xp, scp, dtype):
        if xp.dtype(dtype).kind == 'u':
            # Unsigned integers make underflows in numpy (eventually seen as
            # overflows in `np.exp(tmp)`)
            pytest.skip()
        x = testing.shaped_random((5, 4), xp, dtype=dtype, scale=8)
        return scp.special.softmax(x)

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=atol, rtol=rtol)
    def test_2dim_float16(self, xp, scp):
        x = testing.shaped_random((5, 4), xp, dtype=xp.float16, scale=8)
        return scp.special.softmax(x)

    @testing.for_all_dtypes(no_float16=True, no_bool=True)
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=atol, rtol=rtol)
    def test_multi_dim(self, xp, scp, dtype):
        if xp.dtype(dtype).kind == 'u':
            # Unsigned integers make underflows in numpy (eventually seen as
            # overflows in `np.exp(tmp)`)
            pytest.skip()
        x = testing.shaped_random((5, 6, 7, 4), xp, dtype=dtype, scale=8)
        return scp.special.softmax(x)

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-4, rtol=1e-4)
    def test_multi_dim_float16(self, xp, scp):
        x = testing.shaped_random((5, 6, 7, 4), xp, dtype=xp.float16, scale=8)
        return scp.special.softmax(x)

    @testing.for_all_dtypes(no_float16=True, no_bool=True)
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=atol, rtol=atol)
    def test_2dim_with_axis(self, xp, scp, dtype):
        if xp.dtype(dtype).kind == 'u':
            # Unsigned integers make underflows in numpy (eventually seen as
            # overflows in `np.exp(tmp)`)
            pytest.skip()
        x = testing.shaped_random((5, 4), xp, dtype=dtype, scale=8)
        return scp.special.softmax(x, axis=1)

    @testing.for_all_dtypes(no_bool=True)
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_empty(self, xp, scp, dtype):
        x = testing.shaped_random((), xp, dtype=dtype)
        return scp.special.softmax(x)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_zeros_ones(self, xp, scp):
        x = xp.array([0.0, 1.0, -1.0, -0.0])
        return scp.special.softmax(x)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_nans_infs(self, xp, scp):
        x = xp.array([cupy.inf, cupy.nan, -cupy.nan, -cupy.inf])
        return scp.special.softmax(x)



import cupyx.scipy.special  # NOQA

import cupy

from cupy import testing
from cupy.testing import numpy_cupy_allclose

try:
    import scipy.special  # NOQA
except ImportError:
    pass


atol = {
    'default': 1e-6,
    cupy.float16: 1e-2,
}
rtol = {
    'default': 1e-6,
    cupy.float16: 1e-2,
}


@testing.with_requires("scipy<1.14")
class Testexprel:

    @testing.for_dtypes("efdFD")
    @numpy_cupy_allclose(scipy_name="scp")
    def test_exprel(self, xp, scp, dtype):
        return scp.special.exprel(-1)

    @testing.for_dtypes("efdFD")
    @numpy_cupy_allclose(scipy_name="scp", atol=atol, rtol=rtol)
    def test_exprel_2(self, xp, scp, dtype):
        return scp.special.exprel(1)

    @testing.for_dtypes("efdFD")
    @numpy_cupy_allclose(scipy_name="scp", atol=atol, rtol=rtol)
    def test_exprel_large_values(self, xp, scp, dtype):
        return scp.special.exprel(720)

    @testing.for_dtypes("efdFD")
    @numpy_cupy_allclose(scipy_name="scp", atol=atol, rtol=rtol)
    def test_exprel_small_value(self, xp, scp, dtype):
        return scp.special.exprel(1e-17)

    @testing.for_dtypes("efdFD")
    @numpy_cupy_allclose(scipy_name="scp", atol=atol, rtol=rtol)
    def test_exprel_zero_values(self, xp, scp, dtype):
        return scp.special.exprel(0.0)

    @testing.for_dtypes("edf")
    @numpy_cupy_allclose(scipy_name="scp", atol=atol, rtol=rtol)
    def test_exprel_array_inputs(self, xp, scp, dtype):
        n = testing.shaped_arange((5, 1, 2), xp, dtype) * 0.001
        return scp.special.exprel(n)

    @testing.for_dtypes("edf")
    @numpy_cupy_allclose(scipy_name="scp", atol=atol, rtol=rtol)
    def test_exprel_array_inputs_2(self, xp, scp, dtype):
        n = testing.shaped_random((5, 3), xp, dtype) * 0.001
        return scp.special.exprel(n)



import cupy


_mls_taps = {2: [1], 3: [2], 4: [3], 5: [3], 6: [5], 7: [6], 8: [7, 6, 1],
             9: [5], 10: [7], 11: [9], 12: [11, 10, 4], 13: [12, 11, 8],
             14: [13, 12, 2], 15: [14], 16: [15, 13, 4], 17: [14],
             18: [11], 19: [18, 17, 14], 20: [17], 21: [19], 22: [21],
             23: [18], 24: [23, 22, 17], 25: [22], 26: [25, 24, 20],
             27: [26, 25, 22], 28: [25], 29: [27], 30: [29, 28, 7],
             31: [28], 32: [31, 30, 10]}


MAX_LEN_SEQ_KERNEL = r"""
#include <cupy/math_constants.h>
#include <cupy/carray.cuh>
#include <cupy/complex.cuh>

extern "C" __global__ void max_len_seq(
        long long length, long long n_taps, int n_state, long long* taps,
        signed char* state, signed char* seq) {

    int idx = blockDim.x * blockIdx.x + threadIdx.x;
    for(long long i = 0; i < length; i++) {
        signed char next_state = state[(idx + 1) % n_state];
        if(idx == n_state - 1) {
            seq[i] = state[0];
            for(int n_tap = 0; n_tap < n_taps; n_tap++) {
                long long tap = taps[n_tap];
                next_state ^= state[tap];
            }
        }
        state[idx] = next_state;
    }
}
"""

_max_len_seq = cupy.RawKernel(MAX_LEN_SEQ_KERNEL, 'max_len_seq')


def max_len_seq(nbits, state=None, length=None, taps=None):
    """
    Maximum length sequence (MLS) generator.

    Parameters
    ----------
    nbits : int
        Number of bits to use. Length of the resulting sequence will
        be ``(2**nbits) - 1``. Note that generating long sequences
        (e.g., greater than ``nbits == 16``) can take a long time.
    state : array_like, optional
        If array, must be of length ``nbits``, and will be cast to binary
        (bool) representation. If None, a seed of ones will be used,
        producing a repeatable representation. If ``state`` is all
        zeros, an error is raised as this is invalid. Default: None.
    length : int, optional
        Number of samples to compute. If None, the entire length
        ``(2**nbits) - 1`` is computed.
    taps : array_like, optional
        Polynomial taps to use (e.g., ``[7, 6, 1]`` for an 8-bit sequence).
        If None, taps will be automatically selected (for up to
        ``nbits == 32``).

    Returns
    -------
    seq : array
        Resulting MLS sequence of 0's and 1's.
    state : array
        The final state of the shift register.

    Notes
    -----
    The algorithm for MLS generation is generically described in:

        https://en.wikipedia.org/wiki/Maximum_length_sequence

    The default values for taps are specifically taken from the first
    option listed for each value of ``nbits`` in:

        https://web.archive.org/web/20181001062252/http://www.newwaveinstruments.com/resources/articles/m_sequence_linear_feedback_shift_register_lfsr.htm

    """  # NOQA
    if taps is None:
        if nbits not in _mls_taps:
            known_taps = cupy.array(list(_mls_taps.keys()))
            raise ValueError('nbits must be between %s and %s if taps is None'
                             % (known_taps.min(), known_taps.max()))
        taps = cupy.array(_mls_taps[nbits], cupy.int64)
    else:
        taps = cupy.unique(cupy.array(taps, cupy.int64))[::-1]
        if cupy.any(taps < 0) or cupy.any(taps > nbits) or taps.size < 1:
            raise ValueError('taps must be non-empty with values between '
                             'zero and nbits (inclusive)')
        taps = cupy.array(taps)  # needed for Cython and Pythran

    n_max = (2 ** nbits) - 1
    if length is None:
        length = n_max
    else:
        length = int(length)
        if length < 0:
            raise ValueError('length must be greater than or equal to 0')

    # We use int8 instead of bool here because NumPy arrays of bools
    # don't seem to work nicely with Cython
    if state is None:
        state = cupy.ones(nbits, dtype=cupy.int8, order='c')
    else:
        # makes a copy if need be, ensuring it's 0's and 1's
        state = cupy.array(state, dtype=bool, order='c').astype(cupy.int8)
    if state.ndim != 1 or state.size != nbits:
        raise ValueError('state must be a 1-D array of size nbits')
    if cupy.all(state == 0):
        raise ValueError('state must not be all zeros')

    seq = cupy.empty(length, dtype=cupy.int8, order='c')
    n_taps = len(taps)

    _max_len_seq((1,), (nbits,), (length, n_taps, nbits, taps, state, seq))
    return seq, state



# mypy: ignore-errors
import numpy

if numpy.__version__ < '2':
    from numpy import AxisError  # NOQA
    from numpy import ComplexWarning  # NOQA
    from numpy import ModuleDeprecationWarning  # NOQA
    from numpy import TooHardError  # NOQA
    from numpy import VisibleDeprecationWarning  # NOQA
    from numpy import RankWarning   # NOQA
else:
    from numpy.exceptions import AxisError  # NOQA
    from numpy.exceptions import ComplexWarning  # NOQA
    from numpy.exceptions import ModuleDeprecationWarning  # NOQA
    from numpy.exceptions import TooHardError  # NOQA
    from numpy.exceptions import VisibleDeprecationWarning  # NOQA
    from numpy.exceptions import RankWarning   # NOQA



import numpy

import cupy


def shape(a):
    """Returns the shape of an array

    Args:
        a (array_like): Input array

    Returns:
        tuple of ints: The elements of the shape tuple give the lengths of the
        corresponding array dimensions.

    """
    if isinstance(a, cupy.ndarray):
        return a.shape
    else:
        return numpy.shape(a)


def reshape(a, newshape, order='C'):
    """Returns an array with new shape and same elements.

    It tries to return a view if possible, otherwise returns a copy.

    Args:
        a (cupy.ndarray): Array to be reshaped.
        newshape (int or tuple of ints): The new shape of the array to return.
            If it is an integer, then it is treated as a tuple of length one.
            It should be compatible with ``a.size``. One of the elements can be
            -1, which is automatically replaced with the appropriate value to
            make the shape compatible with ``a.size``.
        order ({'C', 'F', 'A'}):
            Read the elements of ``a`` using this index order, and place the
            elements into the reshaped array using this index order.
            'C' means to read / write the elements using C-like index order,
            with the last axis index changing fastest, back to the first axis
            index changing slowest. 'F' means to read / write the elements
            using Fortran-like index order, with the first index changing
            fastest, and the last index changing slowest. Note that the 'C'
            and 'F' options take no account of the memory layout of the
            underlying array, and only refer to the order of indexing. 'A'
            means to read / write the elements in Fortran-like index order if
            a is Fortran contiguous in memory, C-like order otherwise.

    Returns:
        cupy.ndarray: A reshaped view of ``a`` if possible, otherwise a copy.

    .. seealso:: :func:`numpy.reshape`

    """
    # TODO(okuta): check type
    return a.reshape(newshape, order=order)


def ravel(a, order='C'):
    """Returns a flattened array.

    It tries to return a view if possible, otherwise returns a copy.

    Args:
        a (cupy.ndarray): Array to be flattened.
        order ({'C', 'F', 'A', 'K'}):
            The elements of ``a`` are read using this index order. 'C' means
            to index the elements in row-major, C-style order,
            with the last axis index changing fastest, back to the first
            axis index changing slowest.  'F' means to index the elements
            in column-major, Fortran-style order, with the
            first index changing fastest, and the last index changing
            slowest. Note that the 'C' and 'F' options take no account of
            the memory layout of the underlying array, and only refer to
            the order of axis indexing.  'A' means to read the elements in
            Fortran-like index order if ``a`` is Fortran *contiguous* in
            memory, C-like order otherwise.  'K' means to read the
            elements in the order they occur in memory, except for
            reversing the data when strides are negative.  By default, 'C'
            index order is used.

    Returns:
        cupy.ndarray: A flattened view of ``a`` if possible, otherwise a copy.

    .. seealso:: :func:`numpy.ravel`

    """
    # TODO(beam2d, grlee77): Support ordering option K
    # TODO(okuta): check type
    return a.ravel(order)



import functools
import warnings

import numpy

import cupy
from cupy.exceptions import RankWarning
import cupyx.scipy.fft


def _wraps_polyroutine(func):
    def _get_coeffs(x):
        if isinstance(x, cupy.poly1d):
            return x._coeffs
        if cupy.isscalar(x):
            return cupy.atleast_1d(x)
        if isinstance(x, cupy.ndarray):
            x = cupy.atleast_1d(x)
            if x.ndim == 1:
                return x
            raise ValueError('Multidimensional inputs are not supported')
        raise TypeError('Unsupported type')

    def wrapper(*args):
        coeffs = [_get_coeffs(x) for x in args]
        out = func(*coeffs)

        if all(not isinstance(x, cupy.poly1d) for x in args):
            return out
        if isinstance(out, cupy.ndarray):
            return cupy.poly1d(out)
        if isinstance(out, tuple):
            return tuple([cupy.poly1d(x) for x in out])
        assert False  # Never reach

    return functools.update_wrapper(wrapper, func)


def poly(seq_of_zeros):
    """Computes the coefficients of a polynomial with the given roots sequence.

    Args:
        seq_of_zeros (cupy.ndarray): a sequence of polynomial roots.

    Returns:
        cupy.ndarray: polynomial coefficients from highest to lowest degree.

    .. warning::

        This function doesn't support general 2d square arrays currently.
        Only complex Hermitian and real symmetric 2d arrays are allowed.

    .. seealso:: :func:`numpy.poly`

    """
    x = seq_of_zeros
    if x.ndim == 2 and x.shape[0] == x.shape[1] and x.shape[0] != 0:
        if cupy.array_equal(x, x.conj().T):
            x = cupy.linalg.eigvalsh(x)
        else:
            raise NotImplementedError('Only complex Hermitian and real '
                                      'symmetric 2d arrays are supported '
                                      'currently')
    elif x.ndim == 1:
        x = x.astype(cupy.mintypecode(x.dtype.char), copy=False)
    else:
        raise ValueError('Input must be 1d or non-empty square 2d array.')

    if x.size == 0:
        return 1.0

    size = 2 ** (x.size - 1).bit_length()
    a = cupy.zeros((size, 2), x.dtype)
    a[:, 0].fill(1)
    cupy.negative(x, out=a[:x.size, 1])
    while size > 1:
        size = size // 2
        a = cupy._math.misc._fft_convolve(a[:size], a[size:], 'full')
    return a[0, :x.size + 1]


@_wraps_polyroutine
def polyadd(a1, a2):
    """Computes the sum of two polynomials.

    Args:
        a1 (scalar, cupy.ndarray or cupy.poly1d): first input polynomial.
        a2 (scalar, cupy.ndarray or cupy.poly1d): second input polynomial.

    Returns:
        cupy.ndarray or cupy.poly1d: The sum of the inputs.

    .. seealso:: :func:`numpy.polyadd`

    """
    if a1.size < a2.size:
        a1, a2 = a2, a1
    out = cupy.pad(a2, (a1.size - a2.size, 0))
    out = out.astype(cupy.result_type(a1, a2), copy=False)
    out += a1
    return out


@_wraps_polyroutine
def polysub(a1, a2):
    """Computes the difference of two polynomials.

    Args:
        a1 (scalar, cupy.ndarray or cupy.poly1d): first input polynomial.
        a2 (scalar, cupy.ndarray or cupy.poly1d): second input polynomial.

    Returns:
        cupy.ndarray or cupy.poly1d: The difference of the inputs.

    .. seealso:: :func:`numpy.polysub`

    """
    if a1.shape[0] <= a2.shape[0]:
        out = cupy.pad(a1, (a2.shape[0] - a1.shape[0], 0))
        out = out.astype(cupy.result_type(a1, a2), copy=False)
        out -= a2
    else:
        out = cupy.pad(a2, (a1.shape[0] - a2.shape[0], 0))
        out = out.astype(cupy.result_type(a1, a2), copy=False)
        out -= 2 * out - a1
    return out


@_wraps_polyroutine
def polymul(a1, a2):
    """Computes the product of two polynomials.

    Args:
        a1 (scalar, cupy.ndarray or cupy.poly1d): first input polynomial.
        a2 (scalar, cupy.ndarray or cupy.poly1d): second input polynomial.

    Returns:
        cupy.ndarray or cupy.poly1d: The product of the inputs.

    .. seealso:: :func:`numpy.polymul`

    """
    a1 = cupy.trim_zeros(a1, trim='f')
    a2 = cupy.trim_zeros(a2, trim='f')
    if a1.size == 0:
        a1 = cupy.array([0.], a1.dtype)
    if a2.size == 0:
        a2 = cupy.array([0.], a2.dtype)
    return cupy.convolve(a1, a2)


def _polypow_direct(x, n):
    if n == 0:
        return 1
    if n == 1:
        return x
    if n % 2 == 0:
        return _polypow(cupy.convolve(x, x), n // 2)
    return cupy.convolve(x, _polypow(cupy.convolve(x, x), (n - 1) // 2))


def _polypow(x, n):
    if n == 0:
        return 1
    if n == 1:
        return x

    method = cupy._math.misc._choose_conv_method(x, x, 'full')

    if method == 'direct':
        return _polypow_direct(x, n)
    elif method == 'fft':
        if x.dtype.kind == 'c':
            fft, ifft = cupy.fft.fft, cupy.fft.ifft
        else:
            fft, ifft = cupy.fft.rfft, cupy.fft.irfft
        out_size = (x.size - 1) * n + 1
        size = cupyx.scipy.fft.next_fast_len(out_size)
        fx = fft(x, size)
        fy = cupy.power(fx, n, fx)
        y = ifft(fy, size)
        return y[:out_size]
    else:
        assert False


def _polyfit_typecast(x):
    if x.dtype.kind == 'c':
        return x.astype(numpy.complex128, copy=False)
    return x.astype(numpy.float64, copy=False)


def polyfit(x, y, deg, rcond=None, full=False, w=None, cov=False):
    """Returns the least squares fit of polynomial of degree deg
    to the data y sampled at x.

    Args:
        x (cupy.ndarray): x-coordinates of the sample points of shape (M,).
        y (cupy.ndarray): y-coordinates of the sample points of shape
            (M,) or (M, K).
        deg (int): degree of the fitting polynomial.
        rcond (float, optional): relative condition number of the fit.
            The default value is ``len(x) * eps``.
        full (bool, optional): indicator of the return value nature.
            When False (default), only the coefficients are returned.
            When True, diagnostic information is also returned.
        w (cupy.ndarray, optional): weights applied to the y-coordinates
            of the sample points of shape (M,).
        cov (bool or str, optional): if given, returns the coefficients
            along with the covariance matrix.

    Returns:
        cupy.ndarray or tuple:
        p (cupy.ndarray of shape (deg + 1,) or (deg + 1, K)):
            Polynomial coefficients from highest to lowest degree
        residuals, rank, singular_values, rcond \
        (cupy.ndarray, int, cupy.ndarray, float):
            Present only if ``full=True``.
            Sum of squared residuals of the least-squares fit,
            rank of the scaled Vandermonde coefficient matrix,
            its singular values, and the specified value of ``rcond``.
        V (cupy.ndarray of shape (M, M) or (M, M, K)):
            Present only if ``full=False`` and ``cov=True``.
            The covariance matrix of the polynomial coefficient estimates.

    .. warning::

        cupy.exceptions.RankWarning: The rank of the coefficient matrix in the
        least-squares fit is deficient. It is raised if ``full=False``.

    .. seealso:: :func:`numpy.polyfit`

    """
    if x.dtype.char == 'e' and y.dtype.kind == 'b':
        raise NotImplementedError('float16 x and bool y are not'
                                  ' currently supported')
    if y.dtype == numpy.float16:
        raise TypeError('float16 y are not supported')

    x = _polyfit_typecast(x)
    y = _polyfit_typecast(y)
    deg = int(deg)

    if deg < 0:
        raise ValueError('expected deg >= 0')
    if x.ndim != 1:
        raise TypeError('expected 1D vector for x')
    if x.size == 0:
        raise TypeError('expected non-empty vector for x')
    if y.ndim < 1 or y.ndim > 2:
        raise TypeError('expected 1D or 2D array for y')
    if x.size != y.shape[0]:
        raise TypeError('expected x and y to have same length')

    lhs = cupy.polynomial.polynomial.polyvander(x, deg)[:, ::-1]
    rhs = y

    if w is not None:
        w = _polyfit_typecast(w)
        if w.ndim != 1:
            raise TypeError('expected a 1-d array for weights')
        if w.size != x.size:
            raise TypeError('expected w and y to have the same length')

        lhs *= w[:, None]
        if rhs.ndim == 2:
            w = w[:, None]
        rhs *= w

    if rcond is None:
        rcond = x.size * cupy.finfo(x.dtype).eps

    scale = cupy.sqrt((cupy.square(lhs)).sum(axis=0))
    lhs /= scale
    c, resids, rank, s = cupy.linalg.lstsq(lhs, rhs, rcond)
    if y.ndim > 1:
        scale = scale.reshape(-1, 1)
    c /= scale

    order = deg + 1
    if rank != order and not full:
        msg = 'Polyfit may be poorly conditioned'
        warnings.warn(msg, RankWarning, stacklevel=4)

    if full:
        if resids.dtype.kind == 'c':
            resids = cupy.absolute(resids)
        return c, resids, rank, s, rcond
    if cov:
        base = cupy.linalg.inv(cupy.dot(lhs.T, lhs))
        base /= cupy.outer(scale, scale)

        if cov == 'unscaled':
            factor = 1
        elif x.size > order:
            factor = resids / (x.size - order)
        else:
            raise ValueError('the number of data points must exceed order'
                             ' to scale the covariance matrix')

        if y.ndim != 1:
            base = base[..., None]
        return c, base * factor

    return c


def polyval(p, x):
    """Evaluates a polynomial at specific values.

    Args:
        p (cupy.ndarray or cupy.poly1d): input polynomial.
        x (scalar, cupy.ndarray): values at which the polynomial
        is evaluated.

    Returns:
        cupy.ndarray or cupy.poly1d: polynomial evaluated at x.

    .. warning::

        This function doesn't currently support poly1d values to evaluate.

    .. seealso:: :func:`numpy.polyval`

    """
    if isinstance(p, cupy.poly1d):
        p = p.coeffs
    if not isinstance(p, cupy.ndarray) or p.ndim == 0:
        raise TypeError('p must be 1d ndarray or poly1d object')
    if p.ndim > 1:
        raise ValueError('p must be 1d array')
    if isinstance(x, cupy.poly1d):
        # TODO(asi1024): Needs performance improvement.
        dtype = numpy.result_type(x.coeffs, 1)
        res = cupy.poly1d(cupy.array([0], dtype=dtype))
        prod = cupy.poly1d(cupy.array([1], dtype=dtype))
        for c in p[::-1]:
            res = res + prod * c
            prod = prod * x
        return res
    dtype = numpy.result_type(p.dtype.type(0), x)
    p = p.astype(dtype, copy=False)
    if p.size == 0:
        return cupy.zeros(x.shape, dtype)
    if dtype == numpy.bool_:
        return p.any() * x + p[-1]
    if not cupy.isscalar(x):
        x = cupy.asarray(x, dtype=dtype)[..., None]
    x = x ** cupy.arange(p.size, dtype=dtype)
    return (p[::-1] * x).sum(axis=-1, dtype=dtype)


def roots(p):
    """Computes the roots of a polynomial with given coefficients.

    Args:
        p (cupy.ndarray or cupy.poly1d): polynomial coefficients.

    Returns:
        cupy.ndarray: polynomial roots.

    .. warning::

        This function doesn't support currently polynomial coefficients
        whose companion matrices are general 2d square arrays. Only those
        with complex Hermitian or real symmetric 2d arrays are allowed.

        The current `cupy.roots` doesn't guarantee the order of results.

    .. seealso:: :func:`numpy.roots`

    """
    if isinstance(p, cupy.poly1d):
        p = p.coeffs
    if p.dtype.kind == 'b':
        raise NotImplementedError('boolean inputs are not supported')
    if p.ndim == 0:
        raise TypeError('0-dimensional input is not allowed')
    if p.size < 2:
        return cupy.array([])
    [p] = cupy.polynomial.polyutils.as_series([p[::-1]])
    if p.size < 2:
        return cupy.array([])
    if p.size == 2:
        out = (-p[0] / p[1])[None]
        if p[0] == 0:
            out = out.real.astype(numpy.float64)
        return out
    cmatrix = cupy.polynomial.polynomial.polycompanion(p)
    # TODO(Dahlia-Chehata): Support after cupy.linalg.eigvals is supported
    if cupy.array_equal(cmatrix, cmatrix.conj().T):
        out = cupy.linalg.eigvalsh(cmatrix)
    else:
        raise NotImplementedError('Only complex Hermitian and real '
                                  'symmetric 2d arrays are supported '
                                  'currently')
    return out.astype(p.dtype)



# Keep in sync with typenames exported in `driver.pxd`.

cdef extern from *:
    ctypedef int Device 'CUdevice'
    ctypedef int Result 'CUresult'

    ctypedef void* Context 'CUcontext'
    ctypedef void* Deviceptr 'CUdeviceptr'
    ctypedef void* Event 'CUevent'
    ctypedef void* Function 'CUfunction'
    ctypedef void* Module 'CUmodule'
    ctypedef void* Stream 'CUstream'
    ctypedef void* LinkState 'CUlinkState'

    ctypedef int CUjit_option 'CUjit_option'
    ctypedef int CUjitInputType 'CUjitInputType'
    ctypedef int CUfunction_attribute 'CUfunction_attribute'

    ctypedef size_t(*CUoccupancyB2DSize)(int)

    # For Texture Reference
    ctypedef void* Array 'CUarray_st*'  # = cupy.cuda.runtime.Array
    ctypedef int Array_format 'CUarray_format'
    ctypedef struct Array_desc 'CUDA_ARRAY_DESCRIPTOR':
        Array_format Format
        size_t Height
        unsigned int NumChannels
        size_t Width
    ctypedef int Address_mode 'CUaddress_mode'
    ctypedef int Filter_mode 'CUfilter_mode'



import math
import os
import warnings

import cython
import numpy

import cupy
from cupy._core._kernel import ElementwiseKernel
from cupy._core._reduction import ReductionKernel
from cupy._core._ufuncs import elementwise_copy
import cupy._core.core as core


from libc.stdint cimport intptr_t

from cupy._core cimport _accelerator
from cupy._core._carray cimport shape_t
from cupy._core._dtype cimport to_cuda_dtype
from cupy._core._scalar cimport get_typename
from cupy._core.core cimport _internal_ascontiguousarray
from cupy._core.core cimport _ndarray_init
from cupy._core.core cimport ascontiguousarray
from cupy._core.core cimport _ndarray_base
from cupy._core cimport _memory_range
from cupy._core cimport _routines_manipulation as _manipulation
from cupy._core cimport _routines_math as _math
from cupy.cuda cimport device
from cupy_backends.cuda.api cimport runtime


cdef extern from '../../cupy_backends/cupy_complex.h':
    ctypedef struct cuComplex 'cuComplex':
        float x, y

    ctypedef struct cuDoubleComplex 'cuDoubleComplex':
        double x, y


cdef list compute_types = [COMPUTE_TYPE_TBD,  # float16
                           COMPUTE_TYPE_TBD,  # float32
                           COMPUTE_TYPE_TBD]  # float64
cdef dict compute_type_str = {
    0: 'COMPUTE_TYPE_TBD',
    1: 'COMPUTE_TYPE_DEFAULT',
    2: 'COMPUTE_TYPE_PEDANTIC',
    3: 'COMPUTE_TYPE_FP16',
    4: 'COMPUTE_TYPE_FP32',
    5: 'COMPUTE_TYPE_FP64',
    6: 'COMPUTE_TYPE_BF16',
    7: 'COMPUTE_TYPE_TF32',
}


cpdef int to_compute_type_index(dtype) except -1:
    cdef str dtype_char = numpy.dtype(dtype).char
    if dtype_char == 'e':
        return 0
    elif dtype_char in 'fF':
        return 1
    elif dtype_char in 'dD':
        return 2
    else:
        raise TypeError('dtype is not supported: {}'.format(dtype))


cpdef set_compute_type(dtype, compute_type):
    global compute_types
    if compute_type in (COMPUTE_TYPE_TBD, COMPUTE_TYPE_DEFAULT,
                        COMPUTE_TYPE_PEDANTIC, COMPUTE_TYPE_FP16,
                        COMPUTE_TYPE_FP32, COMPUTE_TYPE_FP64):
        compute_types[to_compute_type_index(dtype)] = compute_type
    elif compute_type in (COMPUTE_TYPE_BF16, COMPUTE_TYPE_TF32):
        if int(device.get_compute_capability()) >= 80:
            compute_types[to_compute_type_index(dtype)] = compute_type
        else:
            warnings.warn('COMPUTE_TYPE_BF16 and COMPUTE_TYPE_TF32 are only '
                          'available on GPUs with compute capability 8.0 or '
                          'higher. COMPUTE_TYPE_DEFAULT will be used instead.')
            compute_types[to_compute_type_index(dtype)] = COMPUTE_TYPE_DEFAULT
    else:
        raise ValueError('Unknown compute type: {}'.format(compute_type))


cpdef compute_type_to_str(compute_type):
    if compute_type in compute_type_str:
        return compute_type_str[compute_type]
    else:
        return compute_type


def _tensordot_core_int_kernel_impl(config, dtype, code, name):
    # This code is based in the GEMM implementation from MAGMA
    # (http://icl.cs.utk.edu/magma/)
    code = '''
#define fetch(arr, col, m, n, bound) arr[min(n*col + m, bound)]

template<typename T>
__device__ void _tensordot_core_int_kernel_impl(
        int M, int N, int K,
        const T* A,
        const T* B,
        T * C)
{
    int idx = threadIdx.x;
    int idy = threadIdx.y;

    int idt = DIM_X * idy + idx;

    int idxA = idt % DIM_XA;
    int idyA = idt / DIM_XA;

    int idxB = idt % DIM_XB;
    int idyB = idt / DIM_XB;

    int blx = blockIdx.x;
    int bly = blockIdx.y;

    __shared__ T sA[BLK_K][BLK_M + 1];
    __shared__ T sB[BLK_N][BLK_K + 1];

    // registers for the innermost loop
    T rC[THR_N][THR_M];
    T rA[THR_M];
    T rB[THR_N];

    T ra[BLK_K / DIM_YA][BLK_M / DIM_XA];
    T rb[BLK_N / DIM_YB][BLK_K / DIM_XB];

    const T* offs_dA = A + blx * BLK_M       + idyA * M + idxA;
    int boundA = (M * (K - 1) + M) - (blx * BLK_M + idyA * M + idxA) - 1;
    const T* offs_dB = B + bly * BLK_N * K + idyB * K + idxB;
    int boundB = (K * (N - 1) + K) - (bly * BLK_N * K + idyB * K + idxB) - 1;

    int m, n, k, kk;

    #pragma unroll
    for (n = 0; n < THR_N; n++) {
        #pragma unroll
        for (m = 0 ; m < THR_M; m++) {
            rC[n][m] = 0;
        }
    }

    // blockwise transpose to transpose load
    #pragma unroll
    for (n = 0; n < BLK_K; n += DIM_YA) {
        #pragma unroll
        for (m = 0; m < BLK_M; m += DIM_XA) {
            sA[n + idyA][m + idxA] = fetch(offs_dA, M, m, n, boundA);
        }
    }
    // blockwise transpose to transpose load
    #pragma unroll
    for (n = 0; n < BLK_N; n += DIM_YB) {
        #pragma unroll
        for (m = 0; m < BLK_K; m += DIM_XB) {
            sB[n + idyB][m + idxB] = fetch(offs_dB, K, m, n, boundB);
        }
    }
    __syncthreads();

    for (kk = 0; kk < K - BLK_K; kk += BLK_K)
    {
        offs_dA += BLK_K * M;
        boundA -= BLK_K * M;
        offs_dB += BLK_K;
        boundB -= BLK_K;

        #pragma unroll
        for (n = 0; n < BLK_K / DIM_YA; n++) {
            #pragma unroll
            for (m = 0; m < BLK_M / DIM_XA; m++) {
                ra[n][m] = fetch(offs_dA, M, m * DIM_XA, n * DIM_YA, boundA);
            }
        }

        #pragma unroll
        for (n = 0; n < BLK_N / DIM_YB; n++) {
            #pragma unroll
            for (m = 0; m < BLK_K / DIM_XB; m++) {
                rb[n][m] = fetch(offs_dB, K, m * DIM_XB, n * DIM_YB, boundB);
            }
        }

        // multiply
        #pragma unroll
        for (k = 0; k < BLK_K; k++)
        {
            #pragma unroll
            for (m = 0; m < THR_M; m++) {
                rA[m] = sA[k][m * DIM_X + idx];
            }

            #pragma unroll
            for (n = 0; n < THR_N; n++) {
                rB[n] = sB[n * DIM_Y + idy][k];
            }

            // HIP is strange...
            #ifdef __HIP_DEVICE_COMPILE__
            __syncthreads();
            #endif

            #pragma unroll
            for (n = 0; n < THR_N; n++) {
                #pragma unroll
                for (m = 0; m < THR_M; m++) {
                    rC[n][m] += rA[m] * rB[n];
                }
            }
        }
        __syncthreads();

        // store A regs->smem
        #pragma unroll
        for (n = 0; n < BLK_K / DIM_YA; n++)
        {
            #pragma unroll
            for (m = 0; m < BLK_M / DIM_XA; m++)
            {
                sA[n * DIM_YA + idyA][m * DIM_XA + idxA] = ra[n][m];
            }
        }

        #pragma unroll
        for (n = 0; n < BLK_N / DIM_YB; n++)
        {
            #pragma unroll
            for (m = 0; m < BLK_K / DIM_XB; m++)
            {
                sB[n * DIM_YB + idyB][m * DIM_XB + idxB] = rb[n][m];
            }
        }
        __syncthreads();
    }

    // Multiply last full (BLK_K) or partial block of columns of A and
    // rows of B.
    // It's okay that m,n exceed matrix bounds as all work is in registers
    // or shared memory, and out-of-bounds rC[n][m] will not be saved later.

    kk = K - kk;
    #pragma unroll
    for (k = 0; k < kk; k++)
    {
        #pragma unroll
        for (m = 0; m < THR_M; m++) {
            rA[m] = sA[k][m * DIM_X + idx];
        }

        #pragma unroll
        for (n = 0; n < THR_N; n++) {
            rB[n] = sB[n * DIM_Y + idy][k];
        }

        // HIP is strange...
        #ifdef __HIP_DEVICE_COMPILE__
        __syncthreads();
        #endif

        #pragma unroll
        for (n = 0; n < THR_N; n++) {
            #pragma unroll
            for (m = 0; m < THR_M; m++) {
                rC[n][m] += rA[m] * rB[n];
            }
        }
    }

    #pragma unroll
    for (n = 0; n < THR_N; n++) {
        int coord_dCn = bly * BLK_N + n * DIM_Y + idy;
        #pragma unroll
        for (m = 0; m < THR_M; m++) {
            int coord_dCm = blx * BLK_M + m * DIM_X + idx;
            if (coord_dCm < M && coord_dCn < N) {
                C[coord_dCn * M + coord_dCm] = rC[n][m];
            }
        }
    }
}
''' + code
    for k, v in config:
        code = '#define ' + k + ' ' + str(v) + '\n' + code
    name_expressions = [f'{name}<bool>',
                        f'{name}<signed char>',
                        f'{name}<unsigned char>',
                        f'{name}<short>',
                        f'{name}<unsigned short>',
                        f'{name}<int>',
                        f'{name}<unsigned int>',
                        f'{name}<long>',
                        f'{name}<unsigned long>',
                        f'{name}<long long>',
                        f'{name}<unsigned long long>']
    mod = cupy.RawModule(code=code, options=('--std=c++11',),
                         name_expressions=name_expressions)
    ker = mod.get_function(name + '<' + get_typename(dtype) + '>')
    return ker


@cupy._util.memoize(for_each_device=True)
def _tensordot_core_int_kernel(config, dtype):
    code = '''
template<typename T>
__global__ void _tensordot_core_int_kernel(
        int M, int N, int K,
        const T* A,
        const T* B,
        T * C)
{
    _tensordot_core_int_kernel_impl(M, N, K, A, B, C);
}
'''
    name = '_tensordot_core_int_kernel'
    return _tensordot_core_int_kernel_impl(config, dtype, code, name)


@cupy._util.memoize(for_each_device=True)
def _tensordot_core_int_batched_kernel(config, dtype):
    code = '''
template<typename T>
__global__ void _tensordot_core_int_batched_kernel(
        int M, int N, int K,
        const T* A[], const T* B[],
        T* C[])
{
    int batchid = blockIdx.z;
    _tensordot_core_int_kernel_impl(
        M, N, K, A[batchid], B[batchid], C[batchid]
    );
}
'''
    name = '_tensordot_core_int_batched_kernel'
    return _tensordot_core_int_kernel_impl(config, dtype, code, name)


@cupy._util.memoize(for_each_device=True)
def _tensordot_core_int_strided_batched_kernel(config, dtype):
    code = '''
template<typename T>
__global__ void _tensordot_core_int_strided_batched_kernel(
        int M, int N, int K,
        const T* A, long long strideA,
        const T* B, long long strideB,
        T * C, long long strideC)
{
    int batchid = blockIdx.z;
    _tensordot_core_int_kernel_impl(
        M, N, K,
        &A[batchid * strideA],
        &B[batchid * strideB],
        &C[batchid * strideC]
    );
}
'''
    name = '_tensordot_core_int_strided_batched_kernel'
    return _tensordot_core_int_kernel_impl(config, dtype, code, name)


cdef tuple _integral_tensordot_core_config():
    # TODO(leofang): autotune the tuning parameters here? See the discussion
    # in this thread: https://groups.google.com/a/icl.utk.edu/g/magma-user/c/igc66uduTfI  # NOQA
    dim_x=16
    dim_y=16
    blk_m=128
    blk_n=128
    blk_k=2
    dim_xa=128
    dim_ya=2
    dim_xb=2
    dim_yb=128
    config = (('DIM_X', dim_x), ('DIM_Y', dim_y),
              ('BLK_M', blk_m), ('BLK_N', blk_n), ('BLK_K', blk_k),
              ('DIM_XA', dim_xa), ('DIM_YA', dim_ya),
              ('DIM_XB', dim_xb), ('DIM_YB', dim_yb),
              ('THR_M', blk_m // dim_x), ('THR_N', blk_n // dim_y))
    return config, dim_x, dim_y, blk_m, blk_n


cdef _ndarray_base _integral_tensordot_core(
        _ndarray_base a, _ndarray_base b, _ndarray_base out, Py_ssize_t m,
        Py_ssize_t n, Py_ssize_t k, str dtype, const shape_t& ret_shape):

    config, dim_x, dim_y, blk_m, blk_n = _integral_tensordot_core_config()
    kern = _tensordot_core_int_kernel(config, dtype)
    args = (m, n, k, a, b, out)
    grid = (int(math.ceil(m / blk_m)), int(math.ceil(n / blk_n)), 1)
    block = (dim_x, dim_y, 1)
    kern(grid, block, args=args)
    return out


cdef _ndarray_base _integral_tensordot_core_batched(
        _ndarray_base a, _ndarray_base b, _ndarray_base out, Py_ssize_t m,
        Py_ssize_t n, Py_ssize_t k, str dtype, Py_ssize_t batch_count):

    config, dim_x, dim_y, blk_m, blk_n = _integral_tensordot_core_config()
    kern = _tensordot_core_int_batched_kernel(config, dtype)
    block = (dim_x, dim_y, 1)
    matPtrA = _mat_ptrs(a)
    matPtrB = _mat_ptrs(b)
    matPtrOut = _mat_ptrs(out)
    max_batch_count = 65000
    for i in range(0, batch_count, max_batch_count):
        ibatch = min(max_batch_count, batch_count - i)
        args = (
            m, n, k, matPtrA[i:i + ibatch], matPtrB[i:i + ibatch],
            matPtrOut[i:i + ibatch])
        grid = (int(math.ceil(m / blk_m)), int(math.ceil(n / blk_n)), ibatch)
        kern(grid, block, args=args)
    return out


cdef _ndarray_base _integral_tensordot_core_strided_batched(
        _ndarray_base a, _ndarray_base b, _ndarray_base out, Py_ssize_t m,
        Py_ssize_t n, Py_ssize_t k, str dtype, Py_ssize_t batch_count):

    config, dim_x, dim_y, blk_m, blk_n = _integral_tensordot_core_config()
    kern = _tensordot_core_int_strided_batched_kernel(config, dtype)
    block = (dim_x, dim_y, 1)
    a = a.reshape((-1,) + a.shape[-2:])
    b = b.reshape((-1,) + b.shape[-2:])
    out = out.reshape((-1,) + out.shape[-2:])
    strideA = _get_stride_for_strided_batched_gemm(a)
    strideB = _get_stride_for_strided_batched_gemm(b)
    strideOut = _get_stride_for_strided_batched_gemm(out)
    max_batch_count = 65000
    for i in range(0, batch_count, max_batch_count):
        ibatch = min(max_batch_count, batch_count - i)
        args = (
            m, n, k, a[i:i + ibatch], strideA, b[i:i + ibatch], strideB,
            out[i:i + ibatch], strideOut)
        grid = (int(math.ceil(m / blk_m)), int(math.ceil(n / blk_n)), ibatch)
        kern(grid, block, args=args)
    return out


cdef _tensordot_core_mul_sum = ReductionKernel(
    'S x, T y', 'U out',
    'static_cast<U>(x) * static_cast<U>(y)',
    'a + b', 'out = a', '0', '_tensordot_core_mul_sum')


cpdef get_compute_type(dtype):
    global compute_types
    cdef int index = to_compute_type_index(dtype)
    if compute_types[index] == COMPUTE_TYPE_TBD:
        compute_type = COMPUTE_TYPE_DEFAULT
        dtype_char = numpy.dtype(dtype).char
        if dtype_char in 'fF' and int(os.getenv('CUPY_TF32', '0')) > 0:
            compute_type = COMPUTE_TYPE_TF32
        set_compute_type(dtype, compute_type)
    return compute_types[index]


@cython.profile(False)
cpdef inline tuple _mat_to_cublas_contiguous(
        _ndarray_base a, Py_ssize_t trans):
    assert a.ndim == 2
    if a._f_contiguous:
        # builtin max function is not used for Cython 0.23
        lda = a._strides[1] // a.itemsize
        if lda < a._shape[0]:
            lda = a._shape[0]
        return a, trans, lda
    if not a._c_contiguous:
        a = a.copy()
    return a, 1 - trans, a._strides[0] // a.itemsize


cpdef _ndarray_base dot(
        _ndarray_base a, _ndarray_base b, _ndarray_base out=None):
    cdef Py_ssize_t a_ndim, b_ndim, a_axis, b_axis, n, m, k
    cdef bint input_a_is_vec, input_b_is_vec
    cdef shape_t ret_shape, shape

    a_ndim = a._shape.size()
    b_ndim = b._shape.size()

    if out is not None:
        if numpy.result_type(a.dtype, b.dtype) != out.dtype:
            raise ValueError('Not supported dtype combination.')
        if not out._c_contiguous:
            raise ValueError('Output array must be C-contiguous')

    if a_ndim == 0 or b_ndim == 0:
        return _math._multiply(a, b, out=out)

    input_a_is_vec = a_ndim == 1
    input_b_is_vec = b_ndim == 1
    if input_a_is_vec:
        shape.clear()
        shape.push_back(1)
        shape.push_back(a.size)
        a = _manipulation._reshape(a, shape)
        a_ndim = 2
    if input_b_is_vec:
        shape.clear()
        shape.push_back(b.size)
        shape.push_back(1)
        b = _manipulation._reshape(b, shape)
        b_ndim = 2

    a_axis = a_ndim - 1
    b_axis = b_ndim - 2

    if a._shape[a_axis] != b._shape[b_axis]:
        raise ValueError('Axis dimension mismatch')

    if a_axis:
        a = _manipulation.rollaxis(a, a_axis, 0)
    if b_axis:
        b = _manipulation.rollaxis(b, b_axis, 0)

    k = a._shape[0]
    if k != 0:
        m = b.size // k
        n = a.size // k
    else:
        # When k==0, the function must return a matrix filled with zero
        # like NumPy.
        m = 0
        n = 0

    if not input_a_is_vec:
        ret_shape.insert(ret_shape.end(), a._shape.begin() + 1, a._shape.end())
    if not input_b_is_vec:
        ret_shape.insert(ret_shape.end(), b._shape.begin() + 1, b._shape.end())
    if out is not None:
        # TODO(kataoka): Make the condition strict
        if k != 0 and out.size != n * m:
            raise ValueError('Output array has an invalid size')

    return tensordot_core(a, b, out, n, m, k, ret_shape)


cpdef _ndarray_base tensordot_core(
        _ndarray_base a, _ndarray_base b, _ndarray_base out, Py_ssize_t n,
        Py_ssize_t m, Py_ssize_t k, const shape_t& ret_shape):
    from cupy_backends.cuda.libs import cublas

    # out, if specified, must be C-contiguous and have correct shape.
    cdef shape_t shape
    cdef Py_ssize_t transa, transb, lda, ldb
    cdef intptr_t handle
    cdef _ndarray_base copy_to_out = None
    cdef str dtype = a.dtype.char
    cdef int compute_capability = int(device.get_compute_capability())
    if dtype != b.dtype.char:
        dtype = numpy.promote_types(dtype, b.dtype).char
    if not a.size or not b.size:
        if out is None:
            out = _ndarray_init(cupy.ndarray, ret_shape, dtype, None)
        out.fill(0)
        return out

    if out is not None:
        assert out.flags.c_contiguous and out.dtype == dtype
    cdef int ace
    if m == 1 and n == 1:
        if out is None:
            out = _ndarray_init(cupy.ndarray, ret_shape, dtype, None)
        c = _manipulation._reshape(out, ())
        for ace in _accelerator._routine_accelerators:
            # fast path using CUB or cuTENSOR
            if ace in (_accelerator.ACCELERATOR_CUB,
                       _accelerator.ACCELERATOR_CUTENSOR):
                (a.ravel() * b.ravel()).sum(out=c)
                break
        else:
            _tensordot_core_mul_sum(a.ravel(), b.ravel(), out=c)
        return out

    a = a.astype(dtype, order='K', casting=None, subok=None, copy=False)
    b = b.astype(dtype, order='K', casting=None, subok=None, copy=False)
    # It copies the operands if needed
    if a._shape.size() != 2 or a._shape[0] != k or a._shape[1] != n:
        shape.clear()
        shape.push_back(k)
        shape.push_back(n)
        a = _manipulation._reshape(a, shape)
    if b._shape.size() != 2 or b._shape[0] != k or b._shape[1] != m:
        shape.clear()
        shape.push_back(k)
        shape.push_back(m)
        b = _manipulation._reshape(b, shape)

    # Be careful that cuBLAS uses the FORTRAN-order matrix representation.
    # Matrix-Matrix product A^T * B
    # c is C-contiguous while cuBLAS assumes F-contiguous inputs, so we
    # compute C^T = B^T * A here.
    a, transa, lda = _mat_to_cublas_contiguous(a, 0)
    b, transb, ldb = _mat_to_cublas_contiguous(b, 1)

    if out is None:
        out = c = _ndarray_init(cupy.ndarray, ret_shape, dtype, None)
    elif (
        _memory_range.may_share_bounds(out, a)
        or _memory_range.may_share_bounds(out, b)
    ):
        copy_to_out = c = _ndarray_init(cupy.ndarray, ret_shape, dtype, None)
    else:
        c = out

    if c._shape.size() != 2 or c._shape[0] != n or c._shape[1] != m:
        c = c.view()
        c.shape = (n, m)

    if dtype not in 'efdFD':
        if transa:
            a = a.T
            a = _internal_ascontiguousarray(a)
        if transb:
            b = _internal_ascontiguousarray(b)
        _integral_tensordot_core(b, a, c, m, n, k, dtype, ret_shape)
        if copy_to_out is not None:
            elementwise_copy(copy_to_out, out)
        return out

    if (
        not runtime._is_hip_environment and
        compute_capability >= 50
    ):
        tensordot_core_v11(transb, transa, m, n, k, b, ldb, a, lda, c, m)
        if copy_to_out is not None:
            elementwise_copy(copy_to_out, out)
        return out

    handle = device.get_cublas_handle()
    if dtype == 'e':
        coef_dtype = 'f'
    else:
        coef_dtype = dtype
    one = numpy.array(1.0, dtype=coef_dtype)
    zero = numpy.array(0.0, dtype=coef_dtype)

    if dtype == 'e':
        use_tensor_core = (not runtime._is_hip_environment and
                           compute_capability >= 70
                           ) or runtime._is_hip_environment
        if use_tensor_core:
            can_opt_in_tensorcore = not runtime._is_hip_environment

            if can_opt_in_tensorcore:
                cublas.setMathMode(handle, cublas.CUBLAS_TENSOR_OP_MATH)
                algo = cublas.CUBLAS_GEMM_DEFAULT_TENSOR_OP
            else:
                algo = cublas.CUBLAS_GEMM_DEFAULT

            cublas.gemmEx(
                handle, <int>transb, <int> transa, <int>m, <int>n, <int>k,
                one.ctypes.data, b.data.ptr, runtime.CUDA_R_16F, <int>ldb,
                a.data.ptr, runtime.CUDA_R_16F, <int>lda, zero.ctypes.data,
                c.data.ptr, runtime.CUDA_R_16F, <int>m, runtime.CUDA_R_32F,
                algo)

            if can_opt_in_tensorcore:
                cublas.setMathMode(handle, cublas.CUBLAS_DEFAULT_MATH)
        else:
            cublas.sgemmEx(
                handle, <int>transb, <int> transa, <int>m, <int>n, <int>k,
                one.ctypes.data, b.data.ptr, runtime.CUDA_R_16F, <int>ldb,
                a.data.ptr, runtime.CUDA_R_16F, <int>lda, zero.ctypes.data,
                c.data.ptr, runtime.CUDA_R_16F, <int>m)
    elif dtype == 'f':
        cublas.sgemmEx(
            handle, <int>transb, <int> transa, <int>m, <int>n, <int>k,
            one.ctypes.data, b.data.ptr, runtime.CUDA_R_32F, <int>ldb,
            a.data.ptr, runtime.CUDA_R_32F, <int>lda, zero.ctypes.data,
            c.data.ptr, runtime.CUDA_R_32F, <int>m)
    elif dtype == 'd':
        cublas.dgemm(
            handle, <int>transb, <int>transa, <int>m, <int>n, <int>k,
            one.ctypes.data, b.data.ptr, <int>ldb, a.data.ptr, <int>lda,
            zero.ctypes.data, c.data.ptr, <int>m)
    elif dtype == 'F':
        cublas.cgemm(
            handle, <int>transb, <int>transa, <int>m, <int>n, <int>k,
            one.ctypes.data, b.data.ptr, <int>ldb, a.data.ptr, <int>lda,
            zero.ctypes.data, c.data.ptr, <int>m)
    elif dtype == 'D':
        cublas.zgemm(
            handle, <int>transb, <int>transa, <int>m, <int>n, <int>k,
            one.ctypes.data, b.data.ptr, <int>ldb, a.data.ptr, <int>lda,
            zero.ctypes.data, c.data.ptr, <int>m)
    else:
        raise ValueError('Invalid dtype: %s' % str(dtype))
    if copy_to_out is not None:
        elementwise_copy(copy_to_out, out)
    return out


cpdef _ndarray_base tensordot_core_v11(
        Py_ssize_t transa, Py_ssize_t transb, Py_ssize_t m, Py_ssize_t n,
        Py_ssize_t k, _ndarray_base a, Py_ssize_t lda, _ndarray_base b,
        Py_ssize_t ldb, _ndarray_base c, Py_ssize_t ldc):
    from cupy_backends.cuda.libs import cublas

    cdef float one_f, zero_f
    cdef double one_d, zero_d
    cdef cuComplex one_F, zero_F
    cdef cuDoubleComplex one_D, zero_D
    cdef size_t one_ptr, zero_ptr

    cdef int compute_capability = int(device.get_compute_capability())
    cdef int compute_type = get_compute_type(c.dtype)
    cdef int cublas_compute_type = -1
    if c.dtype.char in 'efF':
        if compute_type == COMPUTE_TYPE_PEDANTIC:
            cublas_compute_type = cublas.CUBLAS_COMPUTE_32F_PEDANTIC
        elif compute_type == COMPUTE_TYPE_TF32 and c.dtype.char in 'fF':
            cublas_compute_type = cublas.CUBLAS_COMPUTE_32F_FAST_TF32
        else:
            cublas_compute_type = cublas.CUBLAS_COMPUTE_32F
    elif c.dtype.char in 'dD':
        if compute_type == COMPUTE_TYPE_PEDANTIC:
            cublas_compute_type = cublas.CUBLAS_COMPUTE_64F_PEDANTIC
        else:
            cublas_compute_type = cublas.CUBLAS_COMPUTE_64F
    else:
        raise ValueError('Invalid dtype: {}'.format(c.dtype))

    cdef int algo = cublas.CUBLAS_GEMM_DEFAULT
    if ((compute_capability >= 80) or
            (compute_capability >= 70 and c.dtype == 'e')):
        algo = cublas.CUBLAS_GEMM_DEFAULT_TENSOR_OP

    if cublas_compute_type in (cublas.CUBLAS_COMPUTE_32F,
                               cublas.CUBLAS_COMPUTE_32F_PEDANTIC,
                               cublas.CUBLAS_COMPUTE_32F_FAST_TF32):
        if c.dtype.char in 'efd':
            one_f = 1
            zero_f = 0
            one_ptr = <size_t>&one_f
            zero_ptr = <size_t>&zero_f
        else:
            one_F = cuComplex(1, 0)
            zero_F = cuComplex(0, 0)
            one_ptr = <size_t>&one_F
            zero_ptr = <size_t>&zero_F
    elif cublas_compute_type in (cublas.CUBLAS_COMPUTE_64F,
                                 cublas.CUBLAS_COMPUTE_64F_PEDANTIC):
        if c.dtype.char in 'efd':
            one_d = 1
            zero_d = 0
            one_ptr = <size_t>&one_d
            zero_ptr = <size_t>&zero_d
        else:
            one_D = cuDoubleComplex(1, 0)
            zero_D = cuDoubleComplex(0, 0)
            one_ptr = <size_t>&one_D
            zero_ptr = <size_t>&zero_D
    else:
        raise ValueError('Invalid cublas compute type: {}'
                         .format(cublas_compute_type))

    cdef int a_cuda_dtype = to_cuda_dtype(a.dtype, is_half_allowed=True)
    cdef int b_cuda_dtype = to_cuda_dtype(b.dtype, is_half_allowed=True)
    cdef int c_cuda_dtype = to_cuda_dtype(c.dtype, is_half_allowed=True)
    cdef intptr_t handle = device.get_cublas_handle()
    cublas.gemmEx(
        handle, <int>transa, <int>transb, <int>m, <int>n, <int>k, one_ptr,
        a.data.ptr, a_cuda_dtype, <int>lda, b.data.ptr, b_cuda_dtype, <int>ldb,
        zero_ptr, c.data.ptr, c_cuda_dtype, <int>ldc, cublas_compute_type,
        algo)


cdef Py_ssize_t _get_stride_for_strided_batched_gemm(
        _ndarray_base a) except? 0:
    cdef int ndim = a._shape.size()
    assert ndim > 2
    assert a._c_contiguous
    return a._shape[ndim - 2] * a._shape[ndim - 1]


cdef _mat_ptrs_kernel = ElementwiseKernel(
    'T base, T stride', 'T out',
    'out = base + _ind.get()[_ind.ndim - 1] * stride', 'cupy_mat_ptrs',
    reduce_dims=False)


cpdef _ndarray_base _mat_ptrs(_ndarray_base a):
    """Creates an array of pointers to matrices
    Args:
        a: A batch of matrices on GPU.
           shape: (A, B, C) -> A ptrs to mat of size (B, C)
           shape: (A_1, ..., A_N, B, C) -> A_1*...*A_N ptrs to mat of
                  size (B, C)
    Returns:
        GPU array of pointers to matrices.
    """
    cdef int ndim = a._shape.size()
    assert ndim > 2
    cdef _ndarray_base idx
    idx = _mat_ptrs_kernel(
        a.data.ptr, a._strides[0],
        core.ndarray((a._shape[0],), dtype=numpy.uintp))

    for i in range(1, ndim - 2):
        idx = _mat_ptrs_kernel(
            idx[:, None], a._strides[i],
            core.ndarray((idx.size, a._shape[i]), dtype=numpy.uintp))
        idx = idx.ravel()
    return idx


cpdef _ndarray_base matmul(
        _ndarray_base a, _ndarray_base b, _ndarray_base out=None):
    """Matrix product of two arrays.

    Returns the matrix product of two arrays and is the implementation of
    the `@` operator introduced in Python 3.5 following PEP465.

    The main difference against cupy.dot are the handling of arrays with more
    than 2 dimensions. For more information see :func:`numpy.matmul`.

    Args:
        a (cupy.ndarray): The left argument.
        b (cupy.ndarray): The right argument.
        out (cupy.ndarray): Output array.

    Returns:
        cupy.ndarray: Output array.

    .. seealso:: :func:`numpy.matmul`

    """
    from cupy_backends.cuda.libs import cublas

    cdef Py_ssize_t i, n, m, ka, kb, a_sh, b_sh, c_sh, ldc
    cdef Py_ssize_t batchCount, a_part_outshape, b_part_outshape
    cdef int orig_a_ndim, orig_b_ndim, a_ndim, b_ndim, ndim
    cdef _ndarray_base ap, bp, cp, c_view
    cdef bint use_broadcast

    orig_a_ndim = a._shape.size()
    orig_b_ndim = b._shape.size()
    if orig_a_ndim == 0 or orig_b_ndim == 0:
        raise ValueError('Scalar operands are not allowed, use \'*\' instead')

    ndim = max(orig_a_ndim, orig_b_ndim)
    if ndim <= 2:
        if out is None:
            return dot(a, b, out)
        ret_dtype = numpy.promote_types(a.dtype, b.dtype)
        if out._c_contiguous and ret_dtype == out.dtype:
            return dot(a, b, out)
        c = _ndarray_init(cupy.ndarray, out._shape, dtype=ret_dtype, obj=None)
        dot(a, b, c)
        elementwise_copy(c, out)
        return out

    orig_a = a
    orig_b = b
    a_part_outshape = b_part_outshape = 0
    if orig_a_ndim == 1:
        a = _manipulation._reshape(a, (1, a.size))
    else:
        a = a.view()
        a_part_outshape = a._shape[orig_a_ndim - 2]
    if orig_b_ndim == 1:
        b = _manipulation._reshape(b, (b.size, 1))
        ldc = 1
    else:
        b = b.view()
        b_part_outshape = ldc = b._shape[orig_b_ndim - 1]

    # expand dims
    a_ndim = a._shape.size()
    b_ndim = b._shape.size()
    if a_ndim < ndim:
        # TODO(niboshi): Confirm update_x_contiguity flags
        a._set_shape_and_strides(
            (1,) * (ndim - a_ndim) + a.shape,
            (0,) * (ndim - a_ndim) + a.strides,
            True, True)
    if b_ndim < ndim:
        # TODO(niboshi): Confirm update_x_contiguity flags
        b._set_shape_and_strides(
            (1,) * (ndim - b_ndim) + b.shape,
            (0,) * (ndim - b_ndim) + b.strides,
            True, True)

    ret_dtype = numpy.promote_types(a.dtype, b.dtype)
    dtype = ret_dtype
    if dtype.char == 'e':
        dtype = numpy.dtype('f')

    a = ascontiguousarray(a, dtype)
    b = ascontiguousarray(b, dtype)

    # broadcast
    batchCount = 1  # batchCount = numpy.prod(out_shape[:-2])
    out_shape = []
    use_broadcast = False
    for i in range(0, ndim - 2):
        a_sh = a._shape[i]
        b_sh = b._shape[i]
        if a_sh != b_sh and a_sh != 1 and b_sh != 1:
            raise ValueError(
                'operands could not be broadcast together with '
                'remapped shapes')

        if a_sh == 0 or b_sh == 0:
            c_sh = 0
        else:
            c_sh = max(a_sh, b_sh)
        batchCount *= c_sh
        out_shape.append(c_sh)
        if a_sh == 1 and c_sh > 1:
            a._strides[i] = 0
            a._shape[i] = c_sh
            a._c_contiguous = a._f_contiguous = False
            use_broadcast = True

        if b_sh == 1 and c_sh > 1:
            b._strides[i] = 0
            b._shape[i] = c_sh
            b._c_contiguous = b._f_contiguous = False
            use_broadcast = True

    if orig_a_ndim != 1:
        out_shape.append(a_part_outshape)
    if orig_b_ndim != 1:
        out_shape.append(b_part_outshape)

    # (A B)^T = B^T A^T
    a, b = b, a

    ka = a._shape[ndim - 2]
    lda = n = a._shape[ndim - 1]
    m = b._shape[ndim - 2]
    ldb = kb = b._shape[ndim - 1]

    if ka != kb:
        raise ValueError(
            'shapes ({}) and ({}) not aligned'.format(
                ','.join([str(_) for _ in orig_a.shape]),
                ','.join([str(_) for _ in orig_b.shape])))

    if out is not None and out.shape != tuple(out_shape):
        raise ValueError('Output array has an invalid size')

    if a.size == 0 or b.size == 0:
        if out is None:
            return cupy.zeros(out_shape, ret_dtype)
        else:
            out.fill(0)
            return out

    if (
        out is not None and out.dtype == dtype and out.flags.c_contiguous
        and not _memory_range.may_share_bounds(out, a)
        and not _memory_range.may_share_bounds(out, b)
    ):
        c = out
    else:
        c = core.ndarray(out_shape, dtype=dtype)
        if out is None:
            if dtype == ret_dtype:
                out = c
            else:
                out = core.ndarray(out_shape, dtype=ret_dtype)

    if orig_a_ndim == 1 or orig_b_ndim == 1:
        c_view = c.view()
        if orig_b_ndim == 1:
            c_view._shape.push_back(1)
            c_view._strides.push_back(0)
        if orig_a_ndim == 1:
            c_view._shape.insert(c_view._shape.end() - 1, 1)
            c_view._strides.insert(c_view._strides.end() - 1, 0)
        assert c_view._c_contiguous
        c_view._update_f_contiguity()
    else:
        c_view = c

    if dtype.char not in 'efdFD':
        if not use_broadcast:
            _integral_tensordot_core_strided_batched(
                a, b, c_view, n, m, ka, dtype.char, batchCount)
        else:
            _integral_tensordot_core_batched(
                a, b, c_view, n, m, ka, dtype.char, batchCount)
        if out is not c:
            elementwise_copy(c, out)
        return out

    cdef intptr_t handle = device.get_cublas_handle()
    cdef int cuda_dtype = to_cuda_dtype(dtype)
    cdef int algo = cublas.CUBLAS_GEMM_DEFAULT

    one = numpy.array(1, dtype=dtype)
    zero = numpy.array(0, dtype=dtype)
    if not use_broadcast:
        strideA = _get_stride_for_strided_batched_gemm(a)
        strideB = _get_stride_for_strided_batched_gemm(b)
        strideC = _get_stride_for_strided_batched_gemm(c_view)
        if dtype.char in 'fFdD':
            cublas.gemmStridedBatchedEx(
                handle,
                0,  # transa
                0,  # transb
                n, m, ka, one.ctypes.data,
                a.data.ptr, cuda_dtype, lda, strideA,
                b.data.ptr, cuda_dtype, ldb, strideB,
                zero.ctypes.data,
                c_view.data.ptr, cuda_dtype, ldc, strideC,
                batchCount, cuda_dtype, algo)
        else:
            raise TypeError(dtype, a.dtype, b.dtype)
    else:
        ap = _mat_ptrs(a)
        bp = _mat_ptrs(b)
        cp = _mat_ptrs(c_view)
        if dtype == numpy.float32:
            cublas.sgemmBatched(
                handle,
                0,  # transa
                0,  # transb
                n, m, ka, one.ctypes.data,
                ap.data.ptr, lda,
                bp.data.ptr, ldb,
                zero.ctypes.data, cp.data.ptr, ldc, batchCount)
        elif dtype == numpy.float64:
            cublas.dgemmBatched(
                handle,
                0,  # transa
                0,  # transb
                n, m, ka, one.ctypes.data,
                ap.data.ptr, lda,
                bp.data.ptr, ldb,
                zero.ctypes.data, cp.data.ptr, ldc, batchCount)
        elif dtype == numpy.complex64:
            cublas.cgemmBatched(
                handle,
                0,  # transa
                0,  # transb
                n, m, ka, one.ctypes.data,
                ap.data.ptr, lda,
                bp.data.ptr, ldb,
                zero.ctypes.data, cp.data.ptr, ldc, batchCount)
        elif dtype == numpy.complex128:
            cublas.zgemmBatched(
                handle,
                0,  # transa
                0,  # transb
                n, m, ka, one.ctypes.data,
                ap.data.ptr, lda,
                bp.data.ptr, ldb,
                zero.ctypes.data, cp.data.ptr, ldc, batchCount)
        else:
            raise TypeError(dtype, a.dtype, b.dtype)

    if out is not c:
        elementwise_copy(c, out)
    return out



from cupyx.scipy.sparse._base import issparse  # NOQA
from cupyx.scipy.sparse._base import isspmatrix  # NOQA
from cupyx.scipy.sparse._base import spmatrix  # NOQA
from cupyx.scipy.sparse._base import SparseWarning  # NOQA
from cupyx.scipy.sparse._base import SparseEfficiencyWarning  # NOQA
from cupyx.scipy.sparse._coo import coo_matrix  # NOQA
from cupyx.scipy.sparse._coo import isspmatrix_coo  # NOQA
from cupyx.scipy.sparse._csc import csc_matrix  # NOQA
from cupyx.scipy.sparse._csc import isspmatrix_csc  # NOQA
from cupyx.scipy.sparse._csr import csr_matrix  # NOQA
from cupyx.scipy.sparse._csr import isspmatrix_csr  # NOQA
from cupyx.scipy.sparse._dia import dia_matrix  # NOQA
from cupyx.scipy.sparse._dia import isspmatrix_dia  # NOQA

from cupyx.scipy.sparse._construct import eye  # NOQA
from cupyx.scipy.sparse._construct import identity  # NOQA
from cupyx.scipy.sparse._construct import rand  # NOQA
from cupyx.scipy.sparse._construct import random  # NOQA
from cupyx.scipy.sparse._construct import spdiags  # NOQA
from cupyx.scipy.sparse._construct import diags  # NOQA

from cupyx.scipy.sparse._construct import bmat  # NOQA
from cupyx.scipy.sparse._construct import hstack  # NOQA
from cupyx.scipy.sparse._construct import vstack  # NOQA

# TODO(unno): implement bsr_matrix
# TODO(unno): implement dok_matrix
# TODO(unno): implement lil_matrix

from cupyx.scipy.sparse._construct import kron  # NOQA
from cupyx.scipy.sparse._construct import kronsum  # NOQA
# TODO(unno): implement diags
# TODO(unno): implement block_diag

from cupyx.scipy.sparse._extract import find  # NOQA
from cupyx.scipy.sparse._extract import tril  # NOQA
from cupyx.scipy.sparse._extract import triu  # NOQA

# TODO(unno): implement save_npz
# TODO(unno): implement load_npz

# TODO(unno): implement isspmatrix_bsr(x)
# TODO(unno): implement isspmatrix_lil(x)
# TODO(unno): implement isspmatrix_dok(x)



# distutils: language = c++

from libc cimport errno
from libc cimport stdlib
from libc.stdint cimport intptr_t
from libc.string cimport memset, memcpy


IF UNAME_SYSNAME == "Windows":
    cdef extern from "stdlib.h" nogil:
        void * _aligned_malloc(size_t size, size_t alignment)
        void _aligned_free(void * memblock)

    cdef inline void * aligned_alloc(size_t alignment, size_t size) nogil:
        return _aligned_malloc(size, alignment)

    cdef inline void aligned_free(void * ptr) nogil:
        _aligned_free(ptr)
ELSE:
    cdef extern from * nogil:
        void * aligned_alloc(size_t alignment, size_t size)

    cdef inline void aligned_free(void * ptr) nogil:
        stdlib.free(ptr)


# CuPy mempool requirement, see ALLOCATION_UNIT_SIZE in cupy/cuda/memory.pyx
DEF ALIGNMENT = 512


cdef public void* _calloc(size_t nmemb, size_t size) nogil:
    errno.errno = 0
    cdef void* buf = aligned_alloc(ALIGNMENT, nmemb * size)
    if buf and errno.errno == 0:
        buf = memset(buf, 0, nmemb * size)

    return buf


cdef public void* _malloc(size_t size) nogil:
    errno.errno = 0
    return aligned_alloc(ALIGNMENT, size)


cdef public void* _realloc(void *ptr, size_t size) nogil:
    errno.errno = 0
    cdef void* buf = stdlib.realloc(ptr, size)
    cdef void* tmp

    if buf and errno.errno == 0 and <intptr_t>(buf) % ALIGNMENT != 0:
        tmp = buf
        errno.errno = 0
        buf = aligned_alloc(ALIGNMENT, size)
        if buf and errno.errno == 0:
            buf = memcpy(buf, tmp, size)
            stdlib.free(tmp)

    return buf


cdef public void _free(void* ptr) nogil:
    aligned_free(ptr)



import unittest

from cupy import testing


class TestArrayContiguity(unittest.TestCase):

    def test_is_contiguous(self):
        a = testing.shaped_arange((2, 3, 4))
        assert a.flags.c_contiguous is True
        b = a.transpose(2, 0, 1)
        assert b.flags.c_contiguous is False
        c = a[::-1]
        assert c.flags.c_contiguous is False
        d = a[:, :, ::2]
        assert d.flags.c_contiguous is False
